"""Query state and runner."""

import asyncio
import hashlib
import json
import logging
import re
import tempfile
import time
from datetime import datetime
from pathlib import Path
from typing import Optional, Protocol

from aio_overpass.error import (
    ClientError,
    QueryLanguageError,
    QueryRejectCause,
    QueryRejectError,
    ResponseError,
)


__docformat__ = "google"
__all__ = (
    "Query",
    "QueryRunner",
    "DefaultQueryRunner",
    "DEFAULT_MAXSIZE",
    "DEFAULT_TIMEOUT",
)


DEFAULT_MAXSIZE = 512
"""Default ``maxsize`` setting in mebibytes"""

DEFAULT_TIMEOUT = 180
"""Default ``timeout`` setting in seconds"""

_COPYRIGHT = "The data included in this document is from www.openstreetmap.org. The data is made available under ODbL."  # noqa: E501
"""This is the same copyright notice included in result sets"""

_SETTING_PATTERN = re.compile(r"\[(\w+?):(.+?)]\s*;?")
"""A pattern to match setting declarations (not the entire settings statement)."""


class Query:
    """
    State of a query that is either pending, successful, or failed.

    Args:
        input_code: The input Overpass QL code. Note that some settings might be changed
                    by query runners, notably the 'timeout' and 'maxsize' settings.
        logger: The logger to use for all logging output related to this query.
        **kwargs: Additional keyword arguments that can be used to identify queries.

    References:
        - https://wiki.openstreetmap.org/wiki/Overpass_API/Overpass_QL
    """

    def __init__(self, input_code: str, logger: Optional[logging.Logger] = None, **kwargs) -> None:
        self._input_code = input_code
        self._logger = logger
        self._kwargs = kwargs

        self._settings = dict(_SETTING_PATTERN.findall(input_code))

        self._settings["out"] = "json"

        if "maxsize" not in self._settings:
            self._settings["maxsize"] = DEFAULT_MAXSIZE * 1024 * 1024

        if "timeout" not in self._settings:
            self._settings["timeout"] = DEFAULT_TIMEOUT

        self._client_timeout: Optional[float] = None

        # set by the client that executes this query
        self._error: Optional[ClientError] = None
        self._response: Optional[dict] = None
        self._response_bytes = 0.0
        self._time_end_try = 0.0  # time the most recent try finished
        self._time_start = 0.0  # time prior to executing the first try
        self._time_start_try = 0.0  # time prior to executing the most recent try
        self._nb_tries = 0

    def _has_cooldown(self) -> bool:
        """When ``True``, we should query the API status to retrieve our cooldown period."""
        return (
            isinstance(self.error, QueryRejectError)
            and self.error.cause == QueryRejectCause.TOO_MANY_QUERIES
        )

    def reset(self) -> None:
        """Reset the query to its initial state, ignoring previous tries."""
        Query.__init__(self, input_code=self._input_code, **self._kwargs)

    @property
    def input_code(self) -> str:
        """The original input Overpass QL source code."""
        return self._input_code

    @property
    def kwargs(self) -> dict:
        """
        Keyword arguments that can be used to identify queries.

        The default query runner will log these values when a query is run.
        """
        return self._kwargs

    @property
    def logger(self) -> Optional[logging.Logger]:
        """If set, this is the logger used for logging output related to this query."""
        return self._logger

    @property
    def nb_tries(self) -> int:
        """Current number of tries."""
        return self._nb_tries

    @property
    def error(self) -> Optional[ClientError]:
        """
        Error of the most recent try.

        Returns:
            an error or ``None`` if the query wasn't tried or hasn't failed
        """
        return self._error

    @property
    def response(self) -> Optional[dict]:
        """The entire JSON response of the query."""
        return self._response

    @property
    def was_cached(self) -> Optional[bool]:
        """
        Indicates whether the query result was cached.

        Returns:
            ``None`` if the query has not been run yet.
            ``True`` if the query has a result set with zero tries.
            ``False`` otherwise.
        """
        if self._response is None:
            return None
        return self._nb_tries == 0

    @property
    def result_set(self) -> Optional[list[dict]]:
        """
        The result set of the query.

        This is open data, licensed under the Open Data Commons Open Database License (ODbL).
        You are free to copy, distribute, transmit and adapt this data, as long as you credit
        OpenStreetMap and its contributors. If you alter or build upon this data, you may
        distribute the result only under the same licence.

        References:
            - https://www.openstreetmap.org/copyright
            - https://opendatacommons.org/licenses/odbl/1-0/
        """
        if not self._response:
            return None
        return self._response["elements"]

    @property
    def response_size_mib(self) -> Optional[float]:
        """The size of the response in mebibytes."""
        if self._response is None:
            return None
        return self._response_bytes / 1024.0 / 1024.0

    @property
    def maxsize_mib(self) -> float:
        """
        The current value of the [maxsize:*] setting in mebibytes.

        This size indicates the maximum allowed memory for the query in bytes RAM on the server,
        as expected by the user. If the query needs more RAM than this value, the server may abort
        the query with a memory exhaustion. The higher this size, the more probably the server
        rejects the query before executing it.
        """
        return float(self._settings["maxsize"]) // 1024.0 // 1024.0

    @maxsize_mib.setter
    def maxsize_mib(self, value: float) -> None:
        if value <= 0.0:
            msg = "maxsize_mib must be > 0.0"
            raise ValueError(msg)
        self._settings["maxsize"] = int(value * 1024.0 * 1024.0)

    @property
    def timeout_secs(self) -> int:
        """
        The current value of the [timeout:*] setting in seconds.

        This duration is the maximum allowed runtime for the query in seconds, as expected by the
        user. If the query runs longer than this time, the server may abort the query. The higher
        this duration, the more probably the server rejects the query before executing it.
        """
        return int(self._settings["timeout"])

    @timeout_secs.setter
    def timeout_secs(self, value: int) -> None:
        if value < 1:
            msg = "timeout_secs must be >= 1"
            raise ValueError(msg)
        self._settings["timeout"] = value

    @property
    def run_timeout_secs(self) -> Optional[float]:
        """
        A limit to ``run_duration_secs``, that cancels the query when exceeded.

        Defaults to no timeout.

        The client will raise a ``QueryCancelledError`` if the timeout is reached.

        Not to be confused with ``timeout_secs``, which is a setting for the Overpass API instance,
        that limits the query execution time. Instead, this value can be used to limit the total
        client-side time spent on this query (see ``Client.run_query``).
        """
        return self._client_timeout

    @run_timeout_secs.setter
    def run_timeout_secs(self, value: Optional[float]) -> None:
        if value is not None and value <= 0.0:
            msg = "run_timeout_secs must be > 0"
            raise ValueError(msg)
        self._client_timeout = value

    @property
    def code(self) -> str:
        """
        The Overpass QL source code of this query.

        This is different from ``input_code`` only when it comes to settings.
        """
        settings = "".join((f"[{k}:{v}]" for k, v in self._settings.items())) + ";"

        # Remove the original settings statement
        code = _SETTING_PATTERN.sub("", self._input_code)

        return f"{settings}\n{code}"

    @property
    def cache_key(self) -> str:
        """
        Hash QL code, and return its digest as hexadecimal string.

        The default query runner uses this as cache key.
        """
        # important: do not use 'self._input_code'.
        # that way the timeout and maxsize settings don't affect the digest
        return hashlib.sha256(self.code.encode("utf-8")).hexdigest()

    @property
    def done(self) -> bool:
        """Returns ``True`` if the result set was received."""
        return self._response is not None

    @property
    def query_duration_secs(self) -> Optional[float]:
        """
        How long it took to fetch the result set in seconds.

        This is the duration starting with the API request, and ending once
        the result is written to this query object. Although it depends on how busy
        the API instance is, this can give some indication of how long a query takes.

        This is ``None`` if there is no result set yet, or when it was cached.
        """
        if self._response is None or self.was_cached:
            return None

        return max(0.0, self._time_end_try - self._time_start_try)

    @property
    def run_duration_secs(self) -> Optional[float]:
        """
        The total required time for this query in seconds (so far).

        This is ``None`` if the query has not been run yet, or when its result was cached.
        """
        if self._time_start == 0.0:
            return None

        end = self._time_end_try if self._time_end_try > 0.0 else asyncio.get_event_loop().time()
        return end - self._time_start

    @property
    def api_version(self) -> Optional[str]:
        """
        The Overpass API version used by the queried instance.

        Returns:
            f.e. ``"Overpass API 0.7.56.8 7d656e78"``.

        References:
            - https://wiki.openstreetmap.org/wiki/Overpass_API/versions
        """
        if self._response is None:
            return None

        return self._response["generator"]

    @property
    def timestamp_osm(self) -> Optional[datetime]:
        """
        All OSM edits that have been uploaded before this date are included.

        It can take a couple of minutes for changes to the database to show up in the
        Overpass API query results.
        """
        if self._response is None:
            return None

        date_str = self._response["osm3s"]["timestamp_osm_base"]
        return datetime.strptime(date_str, "%Y-%m-%dT%H:%M:%SZ")

    @property
    def timestamp_areas(self) -> Optional[datetime]:
        """
        All area data edits that have been uploaded before this date are included.

        If the query involves area data processing, this is the date of the latest edit
        that has been considered in the most recent batch run of the area generation.
        Otherwise, it is set to ``None``.
        """
        if self._response is None:
            return None

        date_str = self._response["osm3s"].get("timestamp_areas_base")
        if not date_str:
            return None

        return datetime.strptime(date_str, "%Y-%m-%dT%H:%M:%SZ")

    @property
    def copyright(self) -> str:
        """A copyright notice that comes with the result set."""
        if self._response is None:
            return _COPYRIGHT

        return self._response["osm3s"].get("copyright") or _COPYRIGHT

    def __str__(self) -> str:
        query = f"query {self.kwargs}" if self.kwargs else "query <no kwargs>"

        size = self.response_size_mib
        time_query = self.query_duration_secs
        time_total = self.run_duration_secs

        if self.nb_tries == 0:
            details = "pending"
        elif self.done:
            if self.nb_tries == 1:
                details = f"{size}mb in {time_query:.01f}s"
            else:
                details = f"{size}mb in {time_query:.01f} ({time_total:.01f})s"
        else:
            t = "try" if self.nb_tries == 1 else "tries"
            details = f"failing after {self.nb_tries} {t}, {time_total:.01f}s"

        return f"{query} ({details})"

    def __repr__(self) -> str:
        cls_name = type(self).__name__

        details = {
            "kwargs": self._kwargs,
            "done": self.done,
        }

        if self.nb_tries == 0 or self.error:
            details["tries"] = self.nb_tries

        if self.error:
            details["error"] = type(self.error).__name__

        if self.done:
            details["result_size"] = f"{self.response_size_mib:.02f}mb"

            if not self.was_cached:
                details["query_duration"] = f"{self.query_duration_secs:.02f}s"

        if self.nb_tries > 0:
            details["run_duration"] = f"{self.run_duration_secs:.02f}s"

        details_str = ", ".join((f"{k}={v!r}" for k, v in details.items()))

        return f"{cls_name}({details_str})"


class QueryRunner(Protocol):
    """
    A query runner is an async function that is called before a client makes an API request.

    Query runners can be used to
     - retry queries when they fail
     - modify queries, f.e. to lower/increase their maxsize/timeout
     - log query results & errors
     - implement caching
    """

    async def __call__(self, query: Query) -> None:
        """Called with the current query state before the client makes an API request."""
        pass


class DefaultQueryRunner(QueryRunner):
    """
    The default query runner.

    This runner…
     - …retries with an increasing back-off period in between tries if the server is too busy
     - …retries and doubles timeout and maxsize settings if they were exceeded
     - …limits the number of tries
     - …optionally caches query results in temp files

    This runner does *not*…
     - …limit the total time a query runs, including retries
     - …lower timeout and maxsize settings if the server rejected a query

    Args:
        max_tries: The maximum number of times a query is tried. (5 by default)
        cache_ttl_secs: Amount of seconds a query's result set is cached for.
                        Set to zero to disable caching. (zero by default)
    """

    def __init__(self, max_tries: int = 5, cache_ttl_secs: int = 0) -> None:
        if max_tries < 1:
            msg = "max_tries must be >= 1"
            raise ValueError(msg)

        if cache_ttl_secs < 0:
            msg = "cache_ttl_secs must be >= 0"
            raise ValueError(msg)

        self._max_tries = max_tries
        self._cache_ttl_secs = cache_ttl_secs

    @classmethod
    def _logger(cls, query: Query) -> logging.Logger:
        return query.logger or logging.getLogger(f"{cls.__module__}.{cls.__name__}")

    def _cache_read(self, query: Query) -> None:
        logger = DefaultQueryRunner._logger(query)

        if not self._cache_ttl_secs:
            logger.debug("caching is disabled")
            return

        now = int(time.time())

        file_name = f"{query.cache_key}.json"
        file_path = Path(tempfile.gettempdir()) / file_name

        if not file_path.exists():
            logger.info("result was not cached")
            logger.debug(f"checked for cache at {file_path}")
            return

        try:
            with open(file_path) as file:
                response = json.load(file)
        except (OSError, json.JSONDecodeError):
            logger.exception(f"failed to read cached {query}")

        if response.get(_EXPIRATION_KEY, 0) <= now:
            logger.info(f"{query} cache expired")
            return

        query._response = response
        logger.info(f"{query} was cached")

    def _cache_write(self, query: Query) -> None:
        logger = DefaultQueryRunner._logger(query)

        if not self._cache_ttl_secs:
            return

        now = int(time.time())
        query._response[_EXPIRATION_KEY] = now + self._cache_ttl_secs

        file_name = f"{query.cache_key}.json"
        file_path = Path(tempfile.gettempdir()) / file_name

        logger.debug(f"caching at {file_path}…")

        try:
            with open(file_path, "w") as file:
                json.dump(query._response, file)
        except OSError:
            logger.exception(f"failed to cache {query}")

    async def __call__(self, query: Query) -> None:
        """Called with the current query state before the client makes an API request."""
        logger = DefaultQueryRunner._logger(query)

        # Check cache ahead of first try
        if query.nb_tries == 0:
            self._cache_read(query)

        # Success or cached
        if query.done:
            if not query.was_cached:
                self._cache_write(query)
            return

        err = query.error

        # Do not retry if we exhausted all tries, or when a retry would not change the result.
        failed = query.nb_tries == self._max_tries or isinstance(
            err, (ResponseError, QueryLanguageError)
        )

        # Exhausted all tries; do not retry.
        if err and failed:
            logger.error(f"give up on {query}", exc_info=err)
            raise err

        if isinstance(err, QueryRejectError):
            # Wait if the server is too busy.
            if err.cause == QueryRejectCause.TOO_BUSY:
                backoff = _fibo_backoff_secs(query.nb_tries)
                logger.info(f"retry {query} in {backoff:.1f}s")
                await asyncio.sleep(backoff)

            # Wait until a slot opens if the rate limit was exceeded.
            elif err.cause == QueryRejectCause.TOO_MANY_QUERIES:
                pass  # let client enforce cooldown

            # Double timeout if exceeded.
            elif err.cause == QueryRejectCause.EXCEEDED_TIMEOUT:
                old = f"{query.timeout_secs:.1f}s"
                query.timeout_secs *= 2
                new = f"{query.timeout_secs:.1f}s"
                logger.info(f"increased [timeout:*] for {query} from {old} to {new}")

            # Double maxsize if exceeded.
            elif err.cause == QueryRejectCause.EXCEEDED_MAXSIZE:
                old = f"{query.maxsize_mib:.1f}mib"
                query.maxsize_mib *= 2
                new = f"{query.maxsize_mib:.1f}mib"
                logger.info(f"increased [maxsize:*] for {query} from {old} to {new}")


_EXPIRATION_KEY = "__expiration__"


def _fibo_backoff_secs(tries: int) -> float:
    """Fibonacci sequence without zero: 1, 1, 2, 3, 5, 8, etc."""
    a, b = 1.0, 1.0

    for _ in range(tries):
        a, b = b, a + b

    return a


def __cache_delete(query: Query) -> None:
    """Clear a response cached by the default runner (only to be used in tests)."""
    file_name = f"{query.cache_key}.json"
    file_path = Path(tempfile.gettempdir()) / file_name
    file_path.unlink(missing_ok=True)


def __cache_expire(query: Query) -> None:
    """Clear a response cached by the default runner (only to be used in tests)."""
    file_name = f"{query.cache_key}.json"
    file_path = Path(tempfile.gettempdir()) / file_name

    with open(file_path) as file:
        response = json.load(file)

    response[_EXPIRATION_KEY] = 0

    with open(file_path, "w") as file:
        json.dump(response, file)



"""Async client for the Overpass API."""
import importlib.metadata


__version__ = importlib.metadata.version("aio-overpass")

# we add this to all modules for pdoc;
# see https://pdoc.dev/docs/pdoc.html#use-numpydoc-or-google-docstrings
__docformat__ = "google"

# we also use __all__ in all modules for pdoc; this lets us control the order
__all__ = (
    "Client",
    "ClientError",
    "Query",
    "client",
    "element",
    "error",
    "pt",
    "pt_ordered",
    "ql",
    "query",
)

from .client import Client
from .error import ClientError
from .query import Query



"""Classes and queries specialized on public transportation routes."""

from collections import Counter
from collections.abc import Generator
from dataclasses import dataclass
from enum import Enum, auto
from typing import Any, Optional, Union, cast

from aio_overpass._dist import fast_distance
from aio_overpass.element import (
    Bbox,
    GeoJsonDict,
    Node,
    OverpassDict,
    Relation,
    Relationship,
    Spatial,
    Way,
    collect_elements,
)
from aio_overpass.ql import one_of_filter, poly_clause
from aio_overpass.query import Query

import shapely.ops
from shapely.geometry import GeometryCollection, Point, Polygon
from shapely.geometry.base import BaseGeometry


__docformat__ = "google"
__all__ = (
    "Route",
    "Stop",
    "Connection",
    "Vehicle",
    "RouteScheme",
    "RouteQuery",
    "SingleRouteQuery",
    "RoutesWithinQuery",
    "collect_routes",
)


class RouteQuery(Query):
    """
    Base class for queries that produce ``Route`` and ``RouteSegment`` objects.

    Be aware that to build full ``RouteSegment`` objects with tags and geometry, this query loads:
     - every route member
     - every stop area any stop on the route is related to
     - every stop in one of those stop areas

    Args:
        input_code: A query that puts the desired routes into the ``.routes`` set,
                     and optionally, its route masters into the ``.masters`` set
                     (for example by recursing up from ``.routes``).
    """

    def __init__(self, input_code: str, **kwargs) -> None:
        input_code = f"""
            {input_code}
            .routes >> -> .route_members;
            way.route_members -> .route_ways;

            (
                node.route_members[highway=bus_stop];
                node.route_members[public_transport];
                way .route_members[public_transport];
                rel .route_members[public_transport];
            ) -> .route_pt_members;

            .route_pt_members <;
            rel._[public_transport=stop_area]->.stop_areas;
            node(r.stop_areas:"stop")[public_transport=stop_position]->.stop_area_stops;

            .masters out;
            .routes out geom;
            .route_ways out tags;
            .route_pt_members out geom;
            .stop_areas out;
            .stop_area_stops out;
        """

        super().__init__(input_code, **kwargs)


class SingleRouteQuery(RouteQuery):
    """
    A query that produces a single ``Route`` object.

    Args:
        relation_id: the desired route's relation ID
    """

    def __init__(self, relation_id: int, **kwargs) -> None:
        self.relation_id = relation_id

        input_code = f"""
            rel({self.relation_id});
            rel._[type=route]->.routes;
            .routes <<;
            rel._[type=route_master]->.masters;
        """

        super().__init__(input_code, **kwargs)


@dataclass
class RoutesWithinQuery(RouteQuery):
    """
    A query that produces ``Route`` objects for any route within the exterior of a polygon.

    Args:
        polygon: Any route that has at least one member element within this shape
                 will be in the result set of this query. Note that the route members
                 are not limited to this polygon - the majority of a route may in fact
                 be outside of it. This shape should be simplified, since a larger number
                 of coordinates on the exterior will slow down the query.
        vehicles: A list of transportation modes to filter by, or an empty list to include
                  routes of any type. A non-empty list will filter routes by the ``route``
                  key.
    """

    def __init__(
        self, polygon: Polygon, vehicles: Optional[list["Vehicle"]] = None, **kwargs
    ) -> None:
        if not vehicles:
            vehicles = list(Vehicle)

        self.polygon = polygon
        self.vehicles = vehicles

        region_clause = poly_clause(self.polygon)
        route_filter = one_of_filter("route", *(v.name.lower() for v in vehicles))
        input_code = f"""
            rel{region_clause}{route_filter}[type=route]->.routes;
            rel{region_clause}{route_filter}[type=route_master]->.masters;
        """

        super().__init__(input_code, **kwargs)


class _RouteRole(Enum):
    """
    A role in a route relation that is relevant to public transportation.

    References:
        - https://taginfo.openstreetmap.org/relations/route#roles
    """

    STOP = auto()
    PLATFORM = auto()
    NONE = auto()


class Connection(Enum):
    """
    Indicates whether you can enter, exit, or do both at a stop on a route.

    References:
        - https://taginfo.openstreetmap.org/relations/route#roles
    """

    ENTRY_AND_EXIT = auto()
    ENTRY_ONLY = auto()
    EXIT_ONLY = auto()

    @property
    def entry_possible(self) -> bool:
        """``True`` if you can enter at this stop on the route."""
        return self != Connection.EXIT_ONLY

    @property
    def exit_possible(self) -> bool:
        """``True`` if you can exit at this stop on the route."""
        return self != Connection.ENTRY_ONLY

    def __repr__(self) -> str:
        return f"{type(self).__name__}.{self.name}"


@dataclass
class Stop(Spatial):
    """
    A stop on a public transportation route.

    Typically, a stop is modelled as two members in a relation: a stop_position node with the
    'stop' role, and a platform with the 'platform' role. These members may be grouped in
    a common stop_area.

    Attributes:
        idx: stop index on the route
        platform: the platform node, way or relation associated with this stop, if any
        stop_position: the stop position node associated with this stop, if any
        stop_coords: a point that, compared to ``stop_position``, is guaranteed to be on the
                     track of the route whenever it is set. Only set if you are collecting
                     ``RouteSegments``.
    """

    idx: int
    platform: Optional[Relationship]
    stop_position: Optional[Relationship]
    stop_coords: Union[Node, Point, None]

    @property
    def name(self) -> Optional[str]:
        """
        This stop's name.

        If platform and stop position names are the same, that name will be returned.
        Otherwise, the most common name out of platform, stop position, and all related stop areas
        will be returned.
        """
        stop_pos_name = self.stop_position.member.tag("name") if self.stop_position else None
        platform_name = self.platform.member.tag("name") if self.platform else None

        if stop_pos_name == platform_name and stop_pos_name is not None:
            return stop_pos_name

        names = [stop_pos_name, platform_name, *(rel.tag("name") for rel in self.stop_areas)]
        names = [name for name in names if name]

        if not names:
            return None

        counter = Counter(names)
        ((most_common, _),) = counter.most_common(1)
        return most_common

    @property
    def connection(self) -> Connection:
        """Indicates whether you can enter, exit, or do both at this stop."""
        options = [
            _connection(relship) for relship in (self.stop_position, self.platform) if relship
        ]
        return next(
            (opt for opt in options if opt != Connection.ENTRY_AND_EXIT), Connection.ENTRY_AND_EXIT
        )

    @property
    def stop_areas(self) -> set[Relation]:
        """Any stop area related to this stop."""
        return {
            relship_to_stop_area.relation
            for relship_to_route in (self.platform, self.stop_position)
            if relship_to_route
            for relship_to_stop_area in relship_to_route.member.relations
            if relship_to_stop_area.relation.tag("public_transport") == "stop_area"
        }

    @property
    def geojson(self) -> GeoJsonDict:
        """A mapping of this object, using the GeoJSON format."""
        # TODO Stop geojson
        raise NotImplementedError

    @property
    def _stop_point(self) -> Optional[Point]:
        """This is set if we have a point that is on the track of the route."""
        if isinstance(self.stop_coords, Node):
            return self.stop_coords.geometry
        if isinstance(self.stop_coords, Point):
            return self.stop_coords
        return None

    @property
    def _geometry(self) -> GeometryCollection:
        """Collection of ``self.platform``, ``self.stop_position`` and  ``self.stop_coords``."""
        geoms = []

        if self.platform:
            # this can be None if the platform is a relation
            geom: Optional[BaseGeometry] = getattr(self.platform.member, "geometry", None)
            if geom:
                geoms.append(geom)

        if self.stop_position:
            member = cast(Node, self.stop_position.member)
            geoms.append(member.geometry)

        if isinstance(self.stop_coords, Point) and self.stop_coords not in geoms:
            geoms.append(self.stop_coords)

        return GeometryCollection(geoms)

    def __repr__(self) -> str:
        if self.stop_position:
            elem = f"stop_position={self.stop_position.member}"
        elif self.platform:
            elem = f"platform={self.platform.member}"
        else:
            raise AssertionError

        return f"{type(self).__name__}({elem}, name='{self.name}')"


class Vehicle(Enum):
    """
    Most common modes of public transportation.

    References:
        - https://wiki.openstreetmap.org/wiki/Relation:route#Public_transport_routes
    """

    # motor vehicles
    BUS = auto()
    TROLLEYBUS = auto()
    MINIBUS = auto()
    SHARE_TAXI = auto()

    # railway vehicles
    TRAIN = auto()
    LIGHT_RAIL = auto()
    SUBWAY = auto()
    TRAM = auto()

    # boats
    FERRY = auto()

    def __repr__(self) -> str:
        return f"{type(self).__name__}.{self.name}"


class RouteScheme(Enum):
    """
    Tagging schemes for public transportation routes.

    References:
        - https://wiki.openstreetmap.org/wiki/Public_transport#Different_tagging_schemas
        - https://wiki.openstreetmap.org/wiki/Key:public_transport:version
    """

    EXPLICIT_V1 = auto()
    EXPLICIT_V2 = auto()

    ASSUME_V1 = auto()
    ASSUME_V2 = auto()

    OTHER = auto()

    @property
    def version_number(self) -> Optional[int]:
        """Public transport tagging scheme."""
        if self in (RouteScheme.EXPLICIT_V1, RouteScheme.ASSUME_V1):
            return 1
        if self in (RouteScheme.EXPLICIT_V2, RouteScheme.ASSUME_V2):
            return 2
        return None

    def __repr__(self) -> str:
        return f"{type(self).__name__}.{self.name}"


@dataclass
class Route(Spatial):
    """
    A public transportation service route, e.g. a bus line.

    Instances of this class are meant to represent OSM routes using the
    'Public Transport Version 2' (PTv2) scheme. Compared to PTv1, this means f.e. that routes
    with multiple directions (forwards, backwards) are represented by multiple route elements.

    Attributes:
        relation: the underlying relation that describes the path taken by a route,
                  and places where people can embark and disembark from the transit service
        scheme: The tagging scheme that was either assumed, or explicitly set on the route relation.
        stops: a sorted list of stops on this route as they would appear on its timetable,
               which was derived from the ``relation``

    References:
        - https://wiki.openstreetmap.org/wiki/Public_transport
        - https://wiki.openstreetmap.org/wiki/Relation:route#Public_transport_routes
    """

    relation: Relation
    scheme: RouteScheme
    stops: list[Stop]

    @property
    def id(self) -> int:
        """Route relation ID."""
        return self.relation.id

    @property
    def tags(self) -> OverpassDict:
        """
        Tags of the route relation.

        Some tags can be inherited from a route master, if not set already.
        """
        from_relation = self.relation.tags or {}
        from_master = {}

        masters = self.masters
        master = masters[0] if len(masters) == 1 else None

        if master and master.tags is not None:
            from_master = {k: v for k, v in master.tags.items() if k in _TAGS_FROM_ROUTE_MASTER}

        return from_master | from_relation

    def tag(self, key: str, default: Any = None) -> Any:
        """
        Get the tag value for the given key.

        Some tags can be inherited from a route master, if not set already.
        """
        value = self.relation.tag(key, default)
        if value is not default:
            return value

        if key not in _TAGS_FROM_ROUTE_MASTER:
            return default

        masters = self.masters
        master = masters[0] if len(masters) == 1 else None

        if not master:
            return default

        return master.tag(key, default)

    @property
    def ways(self) -> list[Way]:
        """
        The ways making up the path of the route.

        Ways may be ordered, and appear multiple times if a way is travelled on more than once.
        All the ways may be contiguous, but gaps are not uncommon.
        """
        return [
            relship.member
            for relship in self.relation.members
            if isinstance(relship.member, Way) and _role(relship) == _RouteRole.NONE
        ]

    @property
    def masters(self) -> list[Relation]:
        """
        Route master relations this route is a part of.

        By convention, this should be a single relation at most.

        References:
            - https://wiki.openstreetmap.org/wiki/Relation:route_master
        """
        return [
            relship.relation
            for relship in self.relation.relations
            if relship.relation.tag("type") == "route_master"
        ]

    @property
    def name_from(self) -> Optional[str]:
        """
        Name of the start station.

        This is either the value of the relation's ``from`` key, a name derived
        from the route's first stop (see ``Stop.name``), or ``None`` if neither
        is available.
        """
        return self.relation.tag(
            key="from",
            default=self.stops[0].name if self.stops else None,
        )

    @property
    def name_to(self) -> Optional[str]:
        """
        Name of the end station.

        This is either the value of the relation's ``to`` key, a name derived
        from the route's last stop (see ``Stop.name``), or ``None`` if neither
        is available.
        """
        return self.relation.tag(
            key="to",
            default=self.stops[-1].name if len(self.stops) > 1 else None,
        )

    @property
    def name_via(self) -> Optional[str]:
        """
        A name of an important station along the route.

        This is either the value of the relation's ``via`` key, or ``None`` if it is not set.
        """
        return self.relation.tag("via")

    @property
    def name(self) -> Optional[str]:
        """
        The name of the route.

        This is either the value relation's ``name`` key, ``{from_} => {via} => {to}`` if at
        least ``from`` and ``to`` are set, or ``None`` otherwise.
        """
        name_tag = self.tag("name")
        if name_tag:
            return name_tag

        from_ = self.name_from
        via = self.name_via
        to = self.name_to

        if from_ and to and via:
            return f"{from_} => {via} => {to}"

        if from_ and to:
            return f"{from_} => {to}"

        return None

    @property
    def vehicle(self) -> Vehicle:
        """
        The mode of transportation used on this route.

        This value corresponds with the value of the relation's ``route`` key.
        """
        if not self.relation.tags or "route" not in self.relation.tags:
            raise AssertionError
        return Vehicle[self.relation.tags["route"].upper()]

    @property
    def bounds(self) -> Optional[Bbox]:
        """The bounding box around all stops of this route."""
        geom = GeometryCollection([stop._geometry for stop in self.stops if stop._geometry])
        return geom.bounds or None

    @property
    def geojson(self) -> GeoJsonDict:
        """A mapping of this object, using the GeoJSON format."""
        # TODO Route geojson
        raise NotImplementedError

    def __repr__(self) -> str:
        return f"{type(self).__name__}(id={self.relation.id}, name='{self.name}')"


_TAGS_FROM_ROUTE_MASTER = {
    "colour",
    "interval",
    "name",
    "network",
    "opening_hours",
    "operator",
    "ref",
    "school",
    "tourism",
    "wheelchair",
}
"""
Tags that, by convention, can be applied to route masters instead of their routes,
in order to avoid duplication.

References:
    - https://wiki.openstreetmap.org/wiki/Relation:route_master
"""


def collect_routes(query: RouteQuery, perimeter: Optional[Polygon] = None) -> list[Route]:
    # TODO the way 'perimeter' works might be confusing
    """
    Consumes the result set of a query and produces ``Route`` objects.

    The order of elements is *not* retained.

    The result set in the input query will be empty after calling this function.

    Args:
        query: The query that produced a result set of route relations.
        perimeter: If set, ``stops`` at the end will be truncated if they are outside of this
                   polygon. This means stops outside the polygon will still be in the list as
                   long as there is at least one following stop that *is* inside the list.
                   Any relation members in ``members`` will not be filtered.

    Raises:
        ValueError: if the input query has no result set

    Returns:
        all routes in the result set of the input query
    """
    elements = collect_elements(query)
    route_rels = [
        cast(Relation, elem)
        for elem in elements
        if isinstance(elem, Relation) and elem.tag("type") == "route"
    ]

    routes = []

    for route_rel in route_rels:
        # Group route relation members per stop:
        # f.e. a platform, and a stop position at the same station.
        stops = list(_stops(route_rel))

        # Filter stops by perimeter: cut off at the last stop that is in the perimeter.
        if perimeter:
            while stops and not perimeter.contains(stops[-1]._stop_point):
                del stops[-1]

        route = Route(
            relation=route_rel,
            scheme=_scheme(route_rel),
            stops=stops,
        )

        routes.append(route)

    return routes


def _scheme(route: Relation) -> RouteScheme:
    """Try to identify a route's tagging scheme."""
    tagged_version = route.tag("public_transport:version")
    if tagged_version == "1":
        return RouteScheme.EXPLICIT_V1
    if tagged_version == "2":
        return RouteScheme.EXPLICIT_V2
    if tagged_version:
        return RouteScheme.OTHER

    # any directed and/or numbered tags like "forward:stop:1" indicate PTv1
    directions = {
        "forward:",
        "forward_",
        "backward:",
        "backward_",
    }
    assume_v1 = any(
        (
            role and (role.startswith(prefix) or role[-1].isnumeric())
            for role, _ in route
            for prefix in directions
        )
    )

    return RouteScheme.ASSUME_V1 if assume_v1 else RouteScheme.ASSUME_V2


def _stops(route_relation: Relation) -> Generator[Stop, None, None]:
    """
    Group route relation members so that each member belongs to the same stop along the route.

    Typically, a stop is represented by a stop position or platform, or both.
    """
    idx = 0

    def to_stop(*selected: Relationship) -> Stop:
        nonlocal idx
        stop_idx = idx
        idx += 1
        return Stop(
            idx=stop_idx,
            platform=next(
                (relship for relship in selected if _role(relship) == _RouteRole.PLATFORM), None
            ),
            stop_position=next(
                (relship for relship in selected if _role(relship) == _RouteRole.STOP), None
            ),
            stop_coords=None,  # set later
        )

    # Consider all route relation members that are tagged or given a role that makes them
    # relevant for public transportation.
    route_members = [
        relship for relship in route_relation.members if _role(relship) is not _RouteRole.NONE
    ]

    # no more than two elements per group (best case: roles "stop" & "platform")
    prev: Optional[Relationship] = None
    for next_ in route_members:
        if not prev:  # case 1: no two members to compare yet
            prev = next_
            continue

        if not _at_same_stop(prev, next_):  # case 2: members are not part of same station
            yield to_stop(prev)
            prev = next_
            continue

        # case 3: members are part of same station
        yield to_stop(prev, next_)
        prev = None

    if prev:
        yield to_stop(prev)


def _connection(relship: Relationship) -> Connection:
    """Returns the connection at the route member, according to its role."""
    if relship.role:
        if relship.role.endswith("_entry_only"):
            return Connection.ENTRY_ONLY

        if relship.role.endswith("_exit_only"):
            return Connection.EXIT_ONLY

    return Connection.ENTRY_AND_EXIT


def _role(relship: Relationship) -> _RouteRole:
    """
    Returns the route member's tagged role, or a fitting fallback.

    If a member is tagged as platform or stop_position, use the role as is.
    Otherwise, assign the roles that fit the member. This means that if there's
    a platform in the route relation, we'll recognize it as such, even if it has
    not been given the platform role (perhaps due to oversight by a contributor).
    """
    if relship.role:
        if relship.role.startswith("platform") or relship.role == "hail_and_ride":
            return _RouteRole.PLATFORM

        if relship.role.startswith("stop"):
            return _RouteRole.STOP

    if (
        relship.member.tag("public_transport") == "platform"
        or relship.member.tag("highway") == "bus_stop"
    ):
        return _RouteRole.PLATFORM

    # The correct tag for this is [public_transport=stop_position], but we'll accept
    # values like [public_transport=station] as well, as long as the member is a node.
    # The assumption is that any node that is not representing a platform is *probably*
    # supposed to represent the stop position.
    if relship.member.type == "node" and relship.member.tag("public_transport"):
        return _RouteRole.STOP

    return _RouteRole.NONE


def _share_stop_area(a: Relationship, b: Relationship) -> bool:
    """``True`` if the given route members share least one common stop area."""
    a_areas = {
        relship.relation.id
        for relship in a.member.relations
        if relship.relation.tag("public_transport") == "stop_area"
    }
    b_areas = {
        relship.relation.id
        for relship in b.member.relations
        if relship.relation.tag("public_transport") == "stop_area"
    }
    return not a_areas.isdisjoint(b_areas)


def _connection_compatible(a: Connection, b: Connection) -> bool:
    """
    Check whether two connections are compatible.

    Returns:
        ``False`` if one of the connections is entry-only, while the other is exit-only,
        ``True`` otherwise.
    """
    return Connection.ENTRY_AND_EXIT in (a, b) or a == b


def _at_same_stop(a: Relationship, b: Relationship) -> bool:
    """
    Check if two members of a route belong to the same stop in the timetable.

    Looking at the route on a map makes it trivial to group elements as part of the same station,
    but automating this requires some heuristics: we will look at members' type, role, name, and
    their proximity to each other.

    If members grouped by this function are far apart, they might be mistagged, i.e. there was a
    mix-up with either the stop position, platform or order (f.e. having "forward" & "backward"
    stops listed consecutively in the route, instead of their correct timetable order).
    """
    # require not both exit_only & enter_only at same stop
    if not _connection_compatible(_connection(a), _connection(b)):
        return False

    # require no duplicate role at same stop
    if _role(a) == _role(b) and _role(a) is not _RouteRole.NONE:
        return False

    # same name, assume same stop
    if a.member.tag("name", 0) == b.member.tag("name", 1):
        return True

    # same stop area, assume same stop
    if _share_stop_area(a, b):
        return True

    # assume same stop if close together
    if not a.member._geometry or not b.member._geometry:
        return False

    pt_a, pt_b = shapely.ops.nearest_points(
        a.member._geometry, b.member._geometry
    )  # euclidean nearest
    distance = fast_distance(*pt_a.coords[0], *pt_b.coords[0])
    return distance <= _MAX_DISTANCE_TO_TRACK


_MAX_DISTANCE_TO_TRACK = 30.0  # meters
"""
An expectation of the maximum distance between a stop position and its platform.

The two should typically be pretty close to each other. Effectively, this value
should be lower than any sensible (beeline) distance between two consecutive
stops on the same route. It should not be used as a hard constraint.
"""



"""Distance functions."""

from math import cos, hypot


def fast_distance(a_lat: float, a_lon: float, b_lat: float, b_lon: float) -> float:
    """
    Distance between two coordinates using equirectangular approximation.

    This uses just one trig and one sqrt function, which makes it very fast.
    It is useful for small distances (up to a few km), but inaccurate for larger
    distances, depending on distance, bearing, and latitude.

    Returns:
        approximate distance in meters between the two coordinates

    References:
        - https://stackoverflow.com/a/53712712
        - https://www.movable-type.co.uk/scripts/latlong.html
        - https://nssdc.gsfc.nasa.gov/planetary/factsheet/earthfact.html
    """
    # Equatorial radius R = 6_378_137m
    # 2*pi*R / 360 = 111319.49079327358m
    # 0.5 * pi/180 = 0.008726646259971648

    x = b_lat - a_lat
    y = (b_lon - a_lon) * cos((b_lat + a_lat) * 0.008726646)
    return 111_319.490793 * hypot(x, y)



"""
Error types.

```
                            (ClientError)
                                  ╷
    ┌──────────────┬──────────────┼────────────┬────────────────┐
    ╵              ╵              ╵            ╵                ╵
CallError     RunnerError   (QueryError)   GiveupError    ResponseError
                                  ╷                             ╷
         ┌────────────────────────┼──────────────────────┐      │
         ╵                        ╵                      ╵      ╵
QueryLanguageError         QueryRejectError          QueryResponseError
```
"""

import html
import re
from dataclasses import dataclass
from enum import Enum, auto
from json import JSONDecodeError
from typing import Optional, Union, no_type_check

import aiohttp
import aiohttp.typedefs


__docformat__ = "google"
__all__ = (
    "ClientError",
    "CallError",
    "ResponseError",
    "GiveupError",
    "QueryError",
    "QueryLanguageError",
    "QueryRejectError",
    "QueryRejectCause",
    "QueryResponseError",
    "RunnerError",
)


class ClientError(Exception):
    """Base exception for failed Overpass API requests and queries."""


@dataclass
class RunnerError(ClientError):
    """
    The query runner raised an exception.

    This is an unexpected error, in contrast to the runner intentionally
    raising ``query.error`` to stop retrying. The cause of this error
    is therefore anything but a ``ClientError``.

    Attributes:
        cause: the exception that caused this error
    """

    cause: BaseException

    def __post_init__(self) -> None:
        # imitate "raise QueryRunnerError(...) from cause"
        self.__cause__ = self.cause

    def __str__(self) -> str:
        return str(self.cause)


@dataclass
class CallError(ClientError):
    """
    Failed to make an API request.

    This error is raised when the client failed to get any response,
    f.e. due to connection issues.

    Attributes:
        cause: the exception that caused this error
    """

    cause: aiohttp.ClientError

    def __post_init__(self) -> None:
        # imitate "raise CallError(...) from cause"
        self.__cause__ = self.cause

    def __str__(self) -> str:
        return str(self.cause)


@dataclass
class ResponseError(ClientError):
    """
    Unexpected API response.

    This error is raised when a request fails, but we can't specifically say why.
    This may indicate a bug on our end, since the client is meant to process almost any
    response of an Overpass API instance.

    Attributes:
        request_info: Contains information about request.
        history: History from failed response.
        status: HTTP status code of response, e.g. ``400``.
        message: Message of response, e.g. ``"OK"``.
        headers: Headers in response, a list of pairs.
    """

    request_info: aiohttp.RequestInfo
    history: tuple[aiohttp.ClientResponse, ...]
    status: int
    message: str
    headers: Optional[aiohttp.typedefs.LooseHeaders]

    @property
    def response(self) -> aiohttp.ClientResponse:
        """Client response returned by ``aiohttp.ClientSession.request()`` and family."""
        return self.history[-1]

    def __str__(self) -> str:
        return (
            f"unexpected response: {self.status}, {self.message!r}, {self.request_info.real_url!r}"
        )


@dataclass
class GiveupError(ClientError):
    """
    The client spent too long running a query, and gave up.

    This error is raised when the run timeout duration set by a query runner is exceeded.

    Attributes:
        kwargs: the query's ``kwargs``
        after_secs: the total time spent on the query
    """

    kwargs: dict
    after_secs: float

    def __str__(self) -> str:
        query = f"query {self.kwargs}" if self.kwargs else "query <no kwargs>"
        return f"gave up on {query} after {self.after_secs:.01f} seconds"


@dataclass
class QueryError(ClientError):
    """
    Base exception for queries that failed at the Overpass API server.

    Attributes:
        kwargs: the query's ``kwargs``
        remarks: the error remarks provided by the API
    """

    kwargs: dict
    remarks: list[str]

    def __str__(self) -> str:
        query = f"query {self.kwargs}" if self.kwargs else "query <no kwargs>"
        first = f"'{self.remarks[0]}'"
        rest = f" (+{len(self.remarks) - 1} more)" if len(self.remarks) > 1 else ""
        return f"{query} failed: {first}{rest}"


@dataclass
class QueryResponseError(ResponseError, QueryError):
    """
    Unexpected query response.

    This error is raised when a query fails (thus extends ``QueryError``),
    but we can't specifically say why (thus also extends ``ResponseError``).
    """

    def __str__(self) -> str:
        return QueryError.__str__(self)


@dataclass
class QueryLanguageError(QueryError):
    """
    Indicates the query's QL code is not valid.

    Retrying is pointless when encountering this error.
    """


class QueryRejectCause(Enum):
    """Details why a query was rejected or cancelled by an API server."""

    TOO_BUSY = auto()
    """
    Gateway rejection. The server has already so much load that the request cannot be executed.

    Smaller ``[timeout:*]`` and/or ``[maxsize:*]`` values might make the request acceptable.
    """

    TOO_MANY_QUERIES = auto()
    """
    Gateway rejection. There are no open slots for queries from your IP address.

    Running queries take up a slot, and the number of slots is limited. A client
    will only run as many concurrent requests as there are slots, which should make this
    a rare error, assuming you are not making requests through another client.
    """

    EXCEEDED_TIMEOUT = auto()
    """
    Runtime rejection. The query has been (or surely will be) running longer than its proposed
    ``[timeout:*]``, and has been cancelled by the server.

    A higher ``[timeout:*]`` value might allow the query run to completion, but also makes it
    more likely to be rejected by a server under heavy load, before executing it (see ``TOO_BUSY``).
    """

    EXCEEDED_MAXSIZE = auto()
    """
    Runtime rejection. The memory required to execute the query has (or surely will) exceed
    its proposed ``[maxsize:*]``, and has been cancelled by the server.

    A higher ``[maxsize:*]`` value might allow the query run to completion, but also makes it
    more likely to be rejected by a server under heavy load, before executing it (see ``TOO_BUSY``).
    """

    def __str__(self) -> str:
        if self == QueryRejectCause.TOO_BUSY:
            return "server too busy"
        if self == QueryRejectCause.TOO_MANY_QUERIES:
            return "too many queries"
        if self == QueryRejectCause.EXCEEDED_TIMEOUT:
            return "exceeded 'timeout'"
        if self == QueryRejectCause.EXCEEDED_MAXSIZE:
            return "exceeded 'maxsize'"
        raise AssertionError()


@dataclass
class QueryRejectError(QueryError):
    """
    A query was rejected or cancelled by the API server.

    Attributes:
        cause: why the query was rejected or cancelled
    """

    cause: QueryRejectCause

    def __str__(self) -> str:
        if self.cause in (QueryRejectCause.TOO_BUSY, QueryRejectCause.TOO_MANY_QUERIES):
            rejection = "query rejected"
        else:
            rejection = "query cancelled"

        return f"{rejection}: {self.cause}"


@no_type_check
def _to_client_error(
    obj: Union[aiohttp.ClientResponse, aiohttp.ClientError]
) -> Union[CallError, ResponseError]:
    """
    Build a ``ClientError`` from either an ``aiohttp`` client error or an unrecognized response.

    Returns:
        - ``CallError`` if ``obj`` is an ``aiohttp.ClientError``,
          but not an ``aiohttp.ClientResponseError``.
        - ``ResponseError`` if ``obj`` is a response or an ``aiohttp.ClientResponseError``.
    """
    if not isinstance(obj, (aiohttp.ClientResponseError, aiohttp.ClientResponse)):
        return CallError(cause=obj)

    error = ResponseError(
        request_info=obj.request_info,
        history=obj.history,
        status=obj.status,
        message=obj.message if isinstance(obj, aiohttp.ClientResponseError) else obj.reason,
        headers=obj.headers,
    )

    if isinstance(obj, aiohttp.ClientResponseError):
        error.__cause__ = obj

    return error


async def _result_or_raise(response: aiohttp.ClientResponse, query_kwargs: dict) -> dict:
    """
    Try to extract the query result set from a response.

    Raises:
        CallError: When there is any sort of connection error.
        ResponseError: When encountering an unexpected response.
        RejectError: When encountering "Too Many Requests" or "Gateway Timeout";
                     when there's a JSON remark indicating query rejection or cancellation;
                     when there's an HTML error message indicating query rejection or cancellation.
        QueryError: When there's any other JSON remark or HTML error message.
    """
    await _raise_for_html_response(response, query_kwargs)

    try:
        json = await response.json()
    except aiohttp.ClientResponseError as err:
        raise _to_client_error(err) from err
    except JSONDecodeError as err:
        raise _response_error(response) from err

    if json is None:
        raise _response_error(response)

    _raise_for_json_remarks(response, json, query_kwargs)

    _raise_for_status(response, query_kwargs)

    return json


def _raise_for_status(response: aiohttp.ClientResponse, query_kwargs: dict) -> None:
    """
    Raise a fitting exception based on the status code.

    Raises:
        RejectError: for status "Too Many Requests" and "Gateway Timeout"
        ResponseError: for any status >= 400

    References:
        - HTTP status codes: http://overpass-api.de/command_line.html
    """
    if response.status == _TOO_MANY_REQUESTS:
        raise QueryRejectError(
            kwargs=query_kwargs, remarks=[], cause=QueryRejectCause.TOO_MANY_QUERIES
        )

    if response.status == _GATEWAY_TIMEOUT:
        raise QueryRejectError(kwargs=query_kwargs, remarks=[], cause=QueryRejectCause.TOO_BUSY)

    if response.status >= _BAD_REQUEST:
        raise _to_client_error(response)


_BAD_REQUEST = 400
_TOO_MANY_REQUESTS = 429
_GATEWAY_TIMEOUT = 504


def _raise_for_json_remarks(
    response: aiohttp.ClientResponse, json: dict, query_kwargs: dict
) -> None:
    """
    Raise a fitting exception based on a remark in a JSON response.

    Raises:
        RejectError: if the API server remarked that the query was rejected or cancelled
        QueryError: if the API server remarked something else
    """
    remark = json.get("remark")
    if not remark:
        return

    timeout_cause = _match_reject_cause(remark)
    if timeout_cause:
        raise QueryRejectError(kwargs=query_kwargs, remarks=[remark], cause=timeout_cause)

    raise _query_response_error(kwargs=query_kwargs, remarks=[remark], response=response)


async def _raise_for_html_response(response: aiohttp.ClientResponse, query_kwargs: dict) -> None:
    """
    Raise a fitting exception based on error remarks in an HTML response.

    Raises:
        RejectError: if one of the error remarks indicate that the query was rejected or cancelled
        QueryError: when encountering other error remarks
        ResponseError: when error remarks cannot be extracted from the response
    """
    if response.content_type != "text/html":
        return

    text = await response.text()

    pattern = re.compile("Error</strong>: (.+?)</p>", re.DOTALL)
    errors = [html.unescape(err.strip()) for err in pattern.findall(text)]

    if not errors:  # unexpected format
        raise _to_client_error(response)

    if any(_is_ql_error(msg) for msg in errors):
        raise QueryLanguageError(kwargs=query_kwargs, remarks=errors)

    reject_causes = [cause for err in errors if (cause := _match_reject_cause(err))]

    if reject_causes:
        raise QueryRejectError(kwargs=query_kwargs, remarks=errors, cause=reject_causes[0])

    raise _query_response_error(kwargs=query_kwargs, remarks=errors, response=response)


def _match_reject_cause(error_msg: str) -> Optional[QueryRejectCause]:
    """
    Check if the given error message indicates that a query was rejected or cancelled.

    AFAIK, neither the 'remarks' in JSON responses, nor the errors listed in HTML responses
    are neatly listed somewhere, but it seems matching a small subset of remarks is enough
    to identify recoverable errors.

    References:
        - Related: https://github.com/DinoTools/python-overpy/issues/62
        - Examples in the API source: https://github.com/drolbr/Overpass-API/search?q=runtime_error
    """
    if "Please check /api/status for the quota of your IP address" in error_msg:
        return QueryRejectCause.TOO_MANY_QUERIES

    if "The server is probably too busy to handle your request" in error_msg:
        return QueryRejectCause.TOO_BUSY

    if "Query timed out" in error_msg:
        return QueryRejectCause.EXCEEDED_TIMEOUT

    if "out of memory" in error_msg:
        return QueryRejectCause.EXCEEDED_MAXSIZE

    return None


def _is_ql_error(message: str) -> bool:
    return "encoding error:" in message or "parse error:" in message or "static error:" in message


def _response_error(response: aiohttp.ClientResponse) -> ResponseError:
    return ResponseError(
        request_info=response.request_info,
        history=response.history,
        status=response.status,
        message=str(response.reason) if response.reason else "",
        headers=response.headers,
    )


def _query_response_error(
    kwargs: dict,
    remarks: list[str],
    response: aiohttp.ClientResponse,
) -> QueryResponseError:
    return QueryResponseError(
        kwargs=kwargs,
        remarks=remarks,
        request_info=response.request_info,
        history=response.history,
        status=response.status,
        message=str(response.reason) if response.reason else "",
        headers=response.headers,
    )



"""Collect the routes of a ``RouteQuery`` with optimized geometry."""

import itertools
from collections.abc import Generator
from dataclasses import dataclass, replace
from typing import Any, Callable, Optional, Union, cast

from aio_overpass._dist import fast_distance
from aio_overpass.element import GeoJsonDict, Node, Relation, Relationship, Spatial, Way

import networkx as nx
import shapely.ops
from networkx import MultiDiGraph
from shapely.geometry import (
    GeometryCollection,
    LineString,
    MultiLineString,
    MultiPoint,
    Point,
    Polygon,
)

from .pt import _MAX_DISTANCE_TO_TRACK, Route, RouteQuery, Stop, collect_routes


__docformat__ = "google"
__all__ = (
    "OrderedRouteView",
    "OrderedRouteViewNode",
    "collect_ordered_routes",
)


@dataclass
class OrderedRouteViewNode:
    """
    A node on the path of a route.

    Attributes:
        lon: Node longitude.
        lat: Node latitude.
        way_id: The ID of the way this node is located on.
                ``None`` if this node is a stop position and there was no
                path to this node.
        path_idx: Index in ``OrderedRouteView.paths`` this node is a part of.
        n_seen_stops: The number of visited stops so far.
                      Increases on every stop (``1..=nb_stops``).
        distance: The approximate distance travelled so far, in meters.
    """

    lon: float
    lat: float
    way_id: Optional[int]
    path_idx: int
    n_seen_stops: int
    distance: float


@dataclass
class OrderedRouteView(Spatial):
    """
    A view of a public transportation route with simplified and directed path geometry.

    Views of a route can start at any stop, and end at a later stop.
    They are continouous if between every stop, there is a path of contigious ways.
    This is not the case for route relations that have a gap in their track.

    Attributes:
        route: The viewed route.
        ordering: The visited nodes while trying to walk the shortest paths between pairs of stops,
                  ordered from start to finish. This is an empty list if this is an empty view.
    """

    route: Route
    ordering: list[OrderedRouteViewNode]

    @property
    def is_continuous(self) -> bool:
        """
        True if there's a path between all stops on the route.

        False for empty views.
        """
        if not self.ordering:
            return False

        grouped = itertools.groupby(iterable=self.ordering, key=lambda node: node.path_idx)

        sequence_actual = [n for n, _grp in grouped]
        sequence_continuous = list(range(min(sequence_actual), max(sequence_actual) + 1))

        # True if 'path_idx' never skips any stop when iterating 'ordering'
        return sequence_actual == sequence_continuous

    @property
    def stops(self) -> list[Stop]:
        """All stops of this view where a path either starts or ends."""
        if not self.ordering:
            return []

        has_path_to_idx = {node.n_seen_stops for node in self.ordering}
        has_path_from_idx = {node.n_seen_stops - 1 for node in self.ordering}
        indexes = has_path_to_idx | has_path_from_idx
        indexes.remove(len(self.route.stops))
        return [self.route.stops[i] for i in sorted(indexes)]

    def _split(
        self, predicate: Callable[[OrderedRouteViewNode, OrderedRouteViewNode], bool]
    ) -> Generator["OrderedRouteView", None, None]:
        if len(self.ordering) < 2:
            return

        ordering: list[OrderedRouteViewNode] = []

        for a, b in zip(self.ordering, self.ordering[1:]):
            if predicate(a, b):
                yield replace(self, ordering=ordering)
                ordering = [a]
            else:
                ordering.append(a)

        ordering.append(b)
        yield replace(self, ordering=ordering)

    def _group_by(self, key: Callable[[OrderedRouteViewNode], Any]) -> list["OrderedRouteView"]:
        return list(self._split(predicate=lambda a, b: key(a) != key(b)))

    def gap_split(self) -> list["OrderedRouteView"]:
        """Split this view wherever there's a gap in between stops."""
        return list(self._split(predicate=lambda a, b: b.path_idx > a.path_idx + 1))

    def stop_split(self) -> list["OrderedRouteView"]:
        """Split this view at every stop, returning views between every pair of stops."""
        return self._group_by(key=lambda node: node.path_idx)

    def take(self, first_n: int) -> "OrderedRouteView":
        """Returns the continuous view that connects a maximum number of stops at the beginning."""
        if first_n < 2:
            msg = "cannot take less than two stops"
            raise ValueError(msg)

        pre_gap, *_ = self.gap_split()

        by_stop = itertools.groupby(pre_gap.ordering, key=lambda node: node.path_idx)
        by_stop_truncated = itertools.islice(by_stop, first_n - 1)

        ordering = [node for _, nodes in by_stop_truncated for node in nodes]

        return replace(self, ordering=ordering)

    def trim(self, distance: float) -> "OrderedRouteView":
        """
        Trim this view to some distance in meters.

        Returns:
            the continuous view that is not longer than a given distance,
            starting from the first stop.
        """
        pre_gap, *_ = self.gap_split()

        distance_start = pre_gap.ordering[0].distance

        by_stop = itertools.groupby(pre_gap.ordering, key=lambda node: node.path_idx)

        ordering: list[OrderedRouteViewNode] = []

        for _, nodes_iter in by_stop:
            nodes = list(nodes_iter)
            distance_end = nodes[-1].distance

            if (distance_end - distance_start) > distance:
                break

            ordering.extend(nodes)

        return replace(self, ordering=ordering)

    @property
    def paths(self) -> list[Optional[LineString]]:
        """
        The simple paths between every pair of stops.

        These linestrings are the shortest paths, merged from contiguous ways in the route relation.
        Whenever there is no path between two stops, a `None` element will be inserted into the
        result list.
        """
        max_nb_paths = len(self.stops) - 1

        grouped = itertools.groupby(iterable=self.ordering, key=lambda node: node.path_idx)

        lines: list[Optional[LineString]] = [None for _ in range(max_nb_paths)]

        for n, nodes_iter in grouped:
            nodes = [(node.lat, node.lon) for node in nodes_iter]
            if len(nodes) < 2:
                continue
            line = LineString(nodes)
            lines[n] = line

        return lines

    @property
    def path(self) -> Union[LineString, MultiLineString]:
        """
        The geometry representing the path travelled on this view from the first to last stop.

        This is the result of ordering (a subset of) ways inside a route relation by the order of
        traversal, and concatenating them. The order is found by building shortest paths between
        stops. Whenever stops cannot be reached with the ways that are included in the relation,
        the geometry will be split up. This happens when there are "holes" in the relation.
        """
        paths = self.paths

        if not any(ls for ls in paths):
            return MultiLineString([])

        # group consecutive stretches of lines or None values
        stretches_grouped = itertools.groupby(iterable=paths, key=bool)

        # select all sets of consecutive lines to merge them
        stretches = (line_strings for has_track, line_strings in stretches_grouped if has_track)

        merged_lines = []

        for line_strings in stretches:
            coords: list[list[float]] = []

            for line in line_strings:
                if not coords:
                    coords.extend(line.coords)
                else:
                    coords.extend(
                        line.coords[1:]
                    )  # ignore first coord, it's equal to the previous one

            merged_line = LineString(coords)
            merged_lines.append(merged_line)

        if len(merged_lines) == 1:
            return merged_lines[0]

        return MultiLineString(merged_lines)

    @property
    def geojson(self) -> GeoJsonDict:
        """A mapping of this object, using the GeoJSON format."""
        # TODO OrderedRouteView geojson
        raise NotImplementedError


def collect_ordered_routes(
    query: RouteQuery, perimeter: Optional[Polygon] = None, n_jobs: int = 1
) -> list[OrderedRouteView]:
    """
    Produce ``OrderedRouteViews`` objects from a result set.

    Compare to ``collect_routes()``, this function tries to build the geometry representing the
    path travelled on every route from the first to last stop. If there are no holes in a route's
    relation, this can typically generate a single line string to represent the entire path of
    a route. Note that routes tagged with the PTv1 scheme will be ignored (in an effort to keep
    the implemenation as simple as possible).

    Since we cannot guarantee that relation members are ordered, we have to convert the route into
    a graph, and find the shortest paths between stops. The more complex the graph, and the more
    stops in a relation, the more expensive it is to generate those paths. Parallelizing this is
    highly recommended, even for a small amount of routes (see ``n_jobs``).

    You should use this instead of ``collect_routes()`` if you are particularly interested in a
    route's path. Here are some example use cases:
     - rendering a route's path between any two stops
     - measuring the route's travelled distance between any two stops
     - validating the order of ways in the relation

    Args:
        query: The query that produced a result set of route relations.
        perimeter: If set, ``stops`` and ``paths`` of resulting routes will be limited
                   to the exterior of this polygon. Any relation members in ``members``
                   will not be filtered.
        n_jobs: The number of workers that generate route geometries in parallel.
                Passing ``-1`` will start a worker for each CPU, ``-2`` will start one for
                each CPU but one, and the default ``1`` will not run any parallel workers at all.

    Raises:
        ValueError: if the input query has no result set
        ImportError: if ``n_jobs != 1`` and the ``joblib`` extra is missing

    Returns:
        all routes in the result set of the input query
    """
    if n_jobs != 1:
        import joblib

    routes = collect_routes(query, perimeter)

    if not routes:
        return []

    if len(routes) == 1:
        n_jobs = 1

    views = []
    views_empty = []
    graphs = []

    for route in routes:
        seg = OrderedRouteView(route=route, ordering=[])

        has_geometry = route.scheme.version_number == 2 and len(route.stops) >= 2

        if not has_geometry:
            views_empty.append(seg)
            continue

        # Idea: when converting the route's track to a directed graph with parallel edges
        # and allowing edges to be traversed only once, the linestring we look for is the
        # one that concatenates the shortest paths between every pair of stops.
        track_graph = _route_graph(route.relation)

        # For each stop, try to find a stop position on the route's graph (its track).
        track_nodes = MultiPoint([Point(*node) for node in track_graph.nodes])
        track_ways = MultiLineString([LineString([u, v]) for u, v, _key in track_graph.edges])
        for stop in route.stops:
            stop.stop_coords = _find_stop_coords(stop, track_graph, track_nodes, track_ways)

        views.append(seg)
        graphs.append(track_graph)

    # Try to find linestrings that connect all pairs of stops.
    if n_jobs == 1:
        for seg, route, graph in zip(views, routes, graphs):
            seg.ordering = _paths(graph, targets=[stop._stop_point for stop in route.stops])
    else:
        # Note: keep in mind that these objects have to be serialized to use in a seperate process,
        # which could take a while for large objects.
        parallel_args = [
            (graph, [stop._stop_point for stop in route.stops])
            for route, graph in zip(views, graphs)
        ]

        # TODO think about using joblib.Parallel's "return_as"
        #   => can produce a generator that yields the results as soon as they are available
        with joblib.parallel_backend(backend="loky", n_jobs=n_jobs):
            paths = joblib.Parallel()(joblib.delayed(_paths)(*args) for args in parallel_args)

        for seg, path in zip(views, paths):
            seg.ordering = path

    return [*views, *views_empty]


def _route_graph(rel: Relation) -> MultiDiGraph:
    """
    Build a directed graph of a route's track.

    In this graph…
     - …nodes will be a tuple of lat, lon
     - …nodes are every node of every way (stop positions can lie anywhere on the track)
     - …ways that are listed more than once in the relation have parallel edges
     - …inverse edges are added for each way, unless it is tagged as a oneway
     - …edges remain unweighted for now
    """
    graph = MultiDiGraph()

    track = [relship for relship in rel.members if _is_track(relship)]

    for relship in track:
        way = cast(Way, relship.member)

        data = {_WAY_ID_KEY: way.id}

        if way.tag("oneway") == "no":
            is_oneway = False
        else:
            is_oneway = any((way.tag(k) in v for k, v in _PT_ONEWAY_TAGS.items()))

        if is_oneway:
            add_forward_edges = relship.role != "backward"
            add_backward_edges = not add_forward_edges
        else:
            add_forward_edges = add_backward_edges = True

        if isinstance(way.geometry, Polygon):
            nodes = list(way.geometry.exterior.coords)
        else:
            nodes = list(way.geometry.coords)

        for a, b in zip(nodes, nodes[1:]):
            if add_forward_edges:
                graph.add_edge(a, b, **data)

            if add_backward_edges:
                graph.add_edge(b, a, **data)

    return graph


_PT_ONEWAY_TAGS = {
    "oneway": {"yes"},
    "highway": {"motorway", "motorway_link", "trunk_link", "primary_link"},
    "junction": {"circular", "roundabout"},
}
"""
Tag values that are commonly associated with oneways.

References:
    - https://wiki.openstreetmap.org/wiki/Key:oneway
    - https://github.com/openstreetmap/iD/blob/develop/modules/osm/tags.js#L81
    - https://wiki.openstreetmap.org/wiki/Talk:Key:oneway
"""


def _is_track(relship: Relationship) -> bool:
    return relship.member.type == "way" and (
        not relship.role or not relship.role.startswith("platform")
    )


def _find_stop_coords(
    stop: Stop, track_graph: MultiDiGraph, track_nodes: MultiPoint, track_ways: MultiLineString
) -> Union[Node, Point, None]:
    """
    Find a node on the track that closesly represents the stop position.

    Args:
        stop: the stop to locate on the graph
        track_graph: the graph that represents the route's track
        track_nodes: a point for every node in the graph
        track_ways: a line string for every edge in the graph

    Returns:
     - None if no appropriate node found
     - Some Node if we found a stop position in either relation or a stop relation
     - Some Point if we found a close node on the track, that is *probably* close to the
       actual stop position
    """
    # (a) check if the route relation has a stop_position for this stop
    if stop.stop_position:
        stop_node = cast(Node, stop.stop_position.member)
        if stop_node.geometry.coords[0] in track_graph:
            return stop_node

    # (b) check if a related stop_area has a stop_position for this stop
    station_stop_positions = (
        cast(Node, member)
        for stop_area in stop.stop_areas
        for _, member in stop_area
        if member.tag("public_transport") == "stop_position"
    )

    stop_pos = next(
        (el for el in station_stop_positions if el.geometry.coords[0] in track_graph), None
    )

    if stop_pos:
        return stop_pos

    # (c) use a node on the graph, that is closest to one of the relation members
    station_geom = GeometryCollection(
        [relship.member._geometry for relship in (stop.stop_position, stop.platform) if relship]
    )

    if not track_nodes or not station_geom:
        return None

    # Calculate the distance of the stop geometry to the track geometry.
    a, b = shapely.ops.nearest_points(track_ways, station_geom)  # euclidean nearest

    distance_to_track = fast_distance(*a.coords[0], *b.coords[0])

    # Idea: if the stop is too far away from the track, it cannot be representative.
    if distance_to_track > _MAX_DISTANCE_TO_TRACK:
        return None

    # Find the node in the graph that is closest to the stop.
    a, _ = shapely.ops.nearest_points(track_nodes, station_geom)  # euclidean nearest

    # This node is *probably* representative of the actual stop position for this stop.
    # It is possible though, that the node is actually close to more than one way
    # on the route's track, and that we choose a node that could be far from the
    # actual stop position.
    return a


def _paths(route_graph: MultiDiGraph, targets: list[Optional[Point]]) -> list[OrderedRouteViewNode]:
    """
    Find shortest paths in the directed route graph between every target stop.

    Edge weights are set to the metric distance between two nodes.

    Not every two stops can be connected, f.e. when they have no representative
    position on the route's track, or when that track has gaps.

    Args:
        route_graph: the unweighted, directed graph of the route's track
        targets: the stop positions to connect
    """
    # set edge weights to metric distance
    for u, v in route_graph.edges():
        if _WEIGHT_KEY in route_graph[u][v][0]:
            continue

        distance = fast_distance(*u, *v)  # meters

        for k in route_graph[u][v]:
            route_graph[u][v][k]["weight"] = distance

        if u not in route_graph[v]:  # if no inverse exists
            continue

        for k in route_graph[v][u]:
            route_graph[v][u][k]["weight"] = distance

    traversal = _Traversal(
        targets_left=targets[1:],
        targets_visited=targets[:1],
        ordering=[],
        distance=0.0,
        path_idx=0,
    )

    traversal = _traverse_graph(graph=route_graph, progress=traversal)

    return traversal.ordering


_GraphNode = tuple[float, float]


@dataclass
class _Traversal:
    ordering: list[OrderedRouteViewNode]
    targets_visited: list[Optional[Point]]
    targets_left: list[Optional[Point]]
    distance: float
    path_idx: int


def _traverse_graph(graph: MultiDiGraph, progress: _Traversal) -> _Traversal:
    """Find shortest paths between targets, while discouraging edges to be traversed twice."""
    if len(progress.targets_left) == 0:
        return progress

    a = progress.targets_visited[-1]
    b = progress.targets_left[0]

    u = a.coords[0] if a else None
    v = b.coords[0] if b else None

    if u != v:
        try:
            path_nodes = nx.shortest_path(graph, source=u, target=v, weight=_WEIGHT_KEY)
            next_progress = _traverse_path(graph, progress, path_nodes)
            return _traverse_graph(graph, next_progress)
        except nx.NetworkXNoPath:
            pass

    next_progress = _Traversal(
        ordering=[
            *progress.ordering,
            OrderedRouteViewNode(
                lon=u[1],
                lat=u[0],
                way_id=None,
                path_idx=progress.path_idx,
                n_seen_stops=len(progress.targets_visited),
                distance=progress.distance,
            ),
        ],
        targets_left=progress.targets_left[1:],
        targets_visited=progress.targets_visited + progress.targets_left[:1],
        distance=progress.distance,
        path_idx=progress.path_idx + 1,
    )
    return _traverse_graph(graph, next_progress)


def _traverse_path(
    graph: MultiDiGraph, progress: _Traversal, path_nodes: list[_GraphNode]
) -> _Traversal:
    """
    Walk the path to visit the next stop, and collect path nodes along the way.

    Repeated traversal of the edge (u, v) is discouraged by setting a large, arbitrary weight.
    Implicitly, this discourages repeated traversal of ways in a route relation. Since the path
    members are supposed to be ordered, ways that are repeatedly traversed should appear more than
    once in the relation. But since we cannot guarantee that, flat-out removal of these edges/ways
    would be too drastic.
    """
    if not path_nodes:
        msg = "expected non-empty list of nodes"
        raise ValueError(msg)

    edges = list(zip(path_nodes, path_nodes[1:]))
    n_seen_stops = len(progress.targets_visited)
    new_ordering = []

    for u, v in edges:
        # don't duplicate last visited stop position node
        if (
            progress.ordering
            and progress.ordering[-1].lat == u[0]
            and progress.ordering[-1].lon == u[1]
        ):
            continue

        # The path does not specify exactly which edge was traversed, so we select
        # the parallel edge of (u, v) that has the smallest weight, and increase it.
        u, v, key, _ = min(
            graph.edges([u, v], keys=True, data=True), key=lambda t: t[3][_WEIGHT_KEY]
        )

        graph[u][v][key][_WEIGHT_KEY] += _WEIGHT_MULTIPLIER

        # Also increase weight of the inverse edge
        if graph.has_edge(v, u, key):
            graph[v][u][key][_WEIGHT_KEY] += _WEIGHT_MULTIPLIER

        way_id = graph[u][v][key][_WAY_ID_KEY]

        way_distance = graph[u][v][key][_WEIGHT_KEY] % _WEIGHT_MULTIPLIER

        new_ordering.append(
            OrderedRouteViewNode(
                lon=u[1],
                lat=u[0],
                way_id=way_id,
                path_idx=progress.path_idx,
                n_seen_stops=n_seen_stops,
                distance=progress.distance,
            )
        )

        progress.distance += way_distance

    new_ordering.append(
        OrderedRouteViewNode(
            lon=v[1],
            lat=v[0],
            way_id=way_id,
            path_idx=progress.path_idx,
            n_seen_stops=n_seen_stops + 1,
            distance=progress.distance,
        )
    )

    return _Traversal(
        ordering=progress.ordering + new_ordering,
        targets_left=progress.targets_left[1:],
        targets_visited=progress.targets_visited + progress.targets_left[:1],
        distance=progress.distance,
        path_idx=progress.path_idx + 1,
    )


_WEIGHT_MULTIPLIER: float = 1e10
_WEIGHT_KEY = "weight"
_WAY_ID_KEY = "way"



"""Overpass QL helpers."""

import itertools

from shapely.geometry import Polygon


def poly_clause(shp: Polygon) -> str:
    """
    Poly region clause.

    This ``(poly:...)`` clause includes results that occur within the exterior of the given
    polygon. The input shape should be simplified, since a larger number of coordinates will
    slow down the query.

    References:
      - https://wiki.openstreetmap.org/wiki/Overpass_API/Language_Guide#Select_region_by_polygon
    """
    flattened_coords = itertools.chain.from_iterable(shp.exterior.coords[:-1])
    bounds = " ".join(map(str, flattened_coords))
    return f'(poly:"{bounds}")'


def one_of_filter(tag: str, *values: str) -> str:
    """
    Tag filter that requires elements to have a tag value that is in the list of the given ones.

    * returns no filter if ``values`` is empty
    * returns a simple ``[key="value1"]`` filter if ``values`` has one item
    * returns a regex filter ``[key~"^value1$|^value2$|..."]`` filter
      if ``values`` has multiple items

    References:
      - https://wiki.openstreetmap.org/wiki/Overpass_API/Language_Guide#Tag_request_clauses_(or_%22tag_filters%22)
    """
    if not values:
        return ""

    if len(values) == 1:
        return f'[{tag}="{values[0]}"]'

    regex = "|".join(f"^{v}$" for v in values)
    return f'[{tag}~"{regex}"]'



"""Typed result set members."""
import math
import re
from abc import ABC, abstractmethod
from collections import defaultdict
from collections.abc import Iterable, Iterator
from dataclasses import dataclass
from typing import Any, Optional, Union, cast

from aio_overpass import Query

import shapely.geometry
import shapely.ops
from shapely.geometry import (
    GeometryCollection,
    LinearRing,
    LineString,
    MultiPolygon,
    Point,
    Polygon,
)
from shapely.geometry.base import BaseGeometry, BaseMultipartGeometry


__docformat__ = "google"
__all__ = (
    "GeoJsonDict",
    "OverpassDict",
    "Bbox",
    "Spatial",
    "Element",
    "Node",
    "Way",
    "Relation",
    "AreaWay",
    "AreaRelation",
    "Relationship",
    "Metadata",
    "collect_elements",
)

GeoJsonDict = dict[str, Any]
"""
A dictionary representing a GeoJSON object.
"""

OverpassDict = dict[str, Any]
"""
A dictionary representing a JSON object returned by the Overpass API.
"""

Bbox = tuple[float, float, float, float]
"""
The bounding box of a spatial object.

This tuple can be understood as any of
    - ``(s, w, n, e)``
    - ``(minlat, minlon, maxlat, maxlon)``
    - ``(minx, miny, maxx, maxy)``
"""


class Spatial(ABC):
    """
    Base class for (groups of) geospatial objects.

    Classes that represent spatial features extend this class and implement the
    ``geojson`` property. Exporting objects in the GeoJSON format should make it possible
    to integrate them with other tools for visualization, or further analysis.

    Objects of this class have the ``__geo_interface__`` property to follow a protocol proposed
    by Sean Gillies, which can make it easier to use spatial data in other Python software
    (https://gist.github.com/sgillies/2217756). An example of this is the ``shape()`` function
    that builds Shapely geometries from any object with the ``__geo_interface__`` property.

    The ability to re-import the exported GeoJSON structures as ``Spatial`` objects is not
    considered here.
    """

    @property
    @abstractmethod
    def geojson(self) -> GeoJsonDict:
        """
        A mapping of this object, using the GeoJSON format.

        The coordinate reference system for all GeoJSON coordinates is ``CRS:84``,
        which means every coordinate is a tuple of longitude and latitude (in that order)
        on the WGS 84 ellipsoid. Note that this order is flipped for all Shapely geometries
        that represent OpenStreetMap elements (latitude first, then longitude).

        References:
            - https://osmdata.openstreetmap.de/info/projections.html
            - https://tools.ietf.org/html/rfc7946#section-4
        """
        raise NotImplementedError

    @property
    def __geo_interface__(self) -> GeoJsonDict:
        """See ``geojson``."""
        return self.geojson


@dataclass(repr=False)
class Metadata:
    """
    Metadata concerning the most recent edit of an OSM element.

    Attributes:
        version: The version number of the element
        timestamp: Timestamp (ISO 8601) of the most recent change of this element
        changeset: The changeset in which the element was most recently changed
        user_name: Name of the user that made the most recent change to the element
        user_id: ID of the user that made the most recent change to the element
    """

    version: int
    timestamp: str
    changeset: int
    user_name: str
    user_id: int


@dataclass(repr=False, eq=False)
class Element(Spatial):
    """
    Elements are the basic components of OpenStreetMap's data.

    A query's result set is made up of these elements.

    Objects of this class do not necessarily describe OSM elements in their entirety.
    The degrees of detail are decided by the ``out`` statements used in an Overpass query:
    using ``out ids`` f.e. would only include an element's ID, but not its tags, geometry, etc.

    Element geometries have coordinates in the EPSG:4326 coordinate reference system,
    meaning that the coordinates are (latitude, longitude) tuples on the WGS 84 reference ellipsoid.
    The geometries are Shapely objects, where the x/y coordinates refer to lat/lon.
    Since Shapely works on the Cartesian plane, not all operations are useful: distances between
    Shapely objects are Euclidean distances for instance - not geodetic distances.

    Tags provide the semantics of elements. There are *classifying* tags, where for a certain key
    there is a limited number of agreed upon values: the ``highway`` tag f.e. is used to identify
    any kind of road, street or path, and to classify its importance within the road network.
    Deviating values are perceived as erroneous. Other tags are *describing*, where any value
    is acceptable for a certain key: the ``name`` tag is a prominent example for this.

    Objects of this class are not meant to represent "derived elements" of the Overpass API,
    that can have an entirely different data structure compared to traditional elements.
    An example of this is the structure produced by ``out count``, but also special statements like
    ``make`` and ``convert``.

    Attributes:
        id: A number that uniquely identifies an element of a certain type
            (nodes, ways and relations each have their own ID space).
            Note that this ID cannot safely be used to link to a specific object on OSM.
            OSM IDs may change at any time, e.g. if an object is deleted and re-added.
        tags: A list of key-value pairs that describe the element, or ``None`` if the element's
              tags not included in the query's result set.
        bounds: The bounding box of this element, or ``None`` when not using ``out bbox``.
                The ``bounds`` property of Shapely geometries can be used as a replacement.
        center: The center of ``bounds``. If you need a coordinate that is inside the element's
                geometry, consider Shapely's ``representative_point()`` and ``centroid``.
        meta: Metadata of this element, or ``None`` when not using ``out meta``
        relations: All relations that are **also in the query's result set**, and that
                   **are known** to contain this element as a member.

    References:
        - https://wiki.openstreetmap.org/wiki/Elements
        - https://wiki.openstreetmap.org/wiki/Map_features
        - https://wiki.openstreetmap.org/wiki/Overpass_API/Overpass_QL#out
        - https://wiki.openstreetmap.org/wiki/Overpass_API/Permanent_ID
    """

    id: int
    tags: Optional[OverpassDict]
    bounds: Optional[Bbox]
    center: Optional[Point]
    meta: Optional[Metadata]
    relations: list["Relationship"]

    def tag(self, key: str, default: Any = None) -> Any:
        """
        Get the tag value for the given key.

        Returns ``default`` if there is no ``key`` tag.

        References:
            - https://wiki.openstreetmap.org/wiki/Tags
        """
        if not self.tags:
            return default
        return self.tags.get(key, default)

    @property
    def type(self) -> str:
        """The element's type: "node", "way", or "relation"."""
        if isinstance(self, Node):
            return "node"
        if isinstance(self, Way):
            return "way"
        if isinstance(self, Relation):
            return "relation"
        raise ValueError()

    @property
    def link(self) -> str:
        """This element on openstreetmap.org."""
        return f"https://www.openstreetmap.org/{self.type}/{self.id}"

    @property
    def wikidata_id(self) -> Optional[str]:
        """
        [Wikidata](https://www.wikidata.org) item ID of this element.

        This is "perhaps, the most stable and reliable manner to obtain Permanent ID
        of relevant spatial features".

        References:
            - https://wiki.openstreetmap.org/wiki/Permanent_ID
            - https://wiki.openstreetmap.org/wiki/Overpass_API/Permanent_ID
            - https://wiki.openstreetmap.org/wiki/Key:wikidata
            - https://www.wikidata.org/wiki/Wikidata:Notability
            - Nodes on Wikidata: https://www.wikidata.org/wiki/Property:P11693
            - Ways on Wikidata: https://www.wikidata.org/wiki/Property:P10689
            - Relations on Wikidata: https://www.wikidata.org/wiki/Property:P402
        """
        # since tag values are not enforced, use a regex to filter out bad IDs
        if self.tags and "wikidata" in self.tags and _WIKIDATA_Q_ID.match(self.tags["wikidata"]):
            return self.tags["wikidata"]
        return None

    @property
    def wikidata_link(self) -> Optional[str]:
        """This element on wikidata.org."""
        if self.wikidata_id:
            return f"https://www.wikidata.org/wiki/{self.wikidata_id}"
        return None

    @property
    def geojson(self) -> GeoJsonDict:
        """
        A mapping of this object, using the GeoJSON format.

        Objects are mapped as the following:
         - ``Node`` -> ``Feature`` with optional ``Point`` geometry
         - ``Way`` -> ``Feature`` with optional ``LineString`` geometry
         - ``AreaWay`` -> ``Feature`` with optional ``Polygon`` geometry
         - ``AreaRelation`` -> ``Feature`` with optional ``Polygon`` or ``MultiPolygon`` geometry
         - ``Relation`` -> ``FeatureCollection`` (nested ``Relations`` are mapped to unlocated
           ``Features``)

        ``Feature`` properties contain all the following keys if they are present for the element:
        ``id``, ``type``, ``role``, ``tags``, ``nodes``, ``bounds``, ``center``, ``timestamp``,
        ``version``, ``changeset``, ``user``, ``uid``.
        The JSON object in ``properties`` is therefore very similar to the original JSON object
        returned by Overpass, skipping only ``geometry``, ``lat`` and ``lon``.

        Additionally, features inside ``FeatureCollections`` receive the special ``__rel__``
        property. Its value is an object containing all properties of the relation the collection
        represents. This works around the fact that ``FeatureCollections`` have no ``properties``
        member. The benefit is that you can take relation tags into consideration when styling
        and rendering their members on a map (f.e. with Leaflet). The downside is that these
        properties are duplicated for every feature.
        """
        if type(self) != Relation:
            return _geojson_feature(self)

        return {
            "type": "FeatureCollection",
            "features": [_geojson_feature(relship) for relship in self.members],
        }

    @property
    def _geometry(self) -> BaseGeometry:
        return getattr(self, "geometry", GeometryCollection())

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.id})"


_WIKIDATA_Q_ID = re.compile(r"^Q\d+$")


def _geojson_properties(obj: Union[Element, "Relationship"]) -> GeoJsonDict:
    elem = obj if isinstance(obj, Element) else obj.member

    properties = {
        "id": elem.id,
        "type": elem.type,
        "tags": elem.tags,
        "bounds": elem.bounds,
        "center": elem.center.coords[0] if elem.center else None,
        "timestamp": elem.meta.timestamp if elem.meta else None,
        "version": elem.meta.version if elem.meta else None,
        "changeset": elem.meta.changeset if elem.meta else None,
        "user": elem.meta.user_name if elem.meta else None,
        "uid": elem.meta.user_id if elem.meta else None,
        "nodes": getattr(elem, "nodes", None),
    }

    properties = {k: v for k, v in properties.items() if v is not None}

    if isinstance(obj, Relationship):
        properties["role"] = obj.role or ""
        properties["__rel__"] = _geojson_properties(obj.relation)

    return properties


def _geojson_geometry(obj: Union[Element, "Relationship"]) -> Optional[GeoJsonDict]:
    elem = obj if isinstance(obj, Element) else obj.member

    geom = elem._geometry
    if not geom:
        return None

    # Flip coordinates for GeoJSON compliance.
    geom = shapely.ops.transform(lambda lat, lon: (lon, lat), geom)

    mapping = shapely.geometry.mapping(geom)
    if mapping["type"] == "LinearRing":  # this geometry does not exist in GeoJSON
        mapping["type"] = "LineString"

    return mapping


def _geojson_bbox(obj: Union[Element, "Relationship"]) -> Optional[Bbox]:
    elem = obj if isinstance(obj, Element) else obj.member

    geom = elem._geometry
    if not geom:
        return None

    bounds = geom.bounds  # can be (nan, nan, nan, nan)
    if not any(math.isnan(c) for c in bounds):
        (minlat, minlon, maxlat, maxlon) = bounds
        return minlon, minlat, maxlon, maxlat

    return None


def _geojson_feature(obj: Union[Element, "Relationship"]) -> GeoJsonDict:
    feature = {
        "type": "Feature",
        "geometry": _geojson_geometry(obj),
        "properties": _geojson_properties(obj),
    }

    bbox = _geojson_bbox(obj)
    if bbox:
        feature["bbox"] = bbox  # type: ignore

    return feature


@dataclass(repr=False, eq=False)
class Node(Element):
    """
    A point in space, at a specific coordinate.

    Nodes are used to define standalone point features (e.g. a bench),
    or to define the shape or "path" of a way.

    Attributes:
        geometry: A Point or ``None`` if the coordinate is not included in the query's result set.

    References:
        - https://wiki.openstreetmap.org/wiki/Node
    """

    geometry: Optional[Point]


@dataclass(repr=False, eq=False)
class Way(Element):
    """
    A way is an ordered list of nodes.

    An open way is a way whose first node is not its last node (e.g. a railway line).
    A closed way is a way whose first node is also its last node, and may be interpreted either
    as a closed polyline (e.g. a roundabout), an area (e.g. a patch of grass), or both
    (e.g. a roundabout surrounding a grassy area).

    For ways that enclose an area, refer to the ``AreaWay`` class.

    Attributes:
        node_ids: The IDs of the nodes that make up this way, or ``None`` if they are not included
                  in the query's result set.
        geometry: A Linestring if the way is open, a LinearRing if the way is closed,
                  a Polygon if the way is closed and its tags indicate that it represents an area,
                  or ``None`` if the geometry is not included in the query's result set.

    References:
        - https://wiki.openstreetmap.org/wiki/Way
    """

    node_ids: Optional[list[int]]
    geometry: Union[LineString, LinearRing, None]


@dataclass(repr=False, eq=False)
class Relation(Element):
    """
    A relation is a group of nodes and ways that have a logical or geographic relationship.

    This relationship is described through its tags.

    For relations that describe an area, refer to the ``AreaRelation`` class.

    Attributes:
        members: Ordered member elements of this relation, with an optional role

    References:
        - https://wiki.openstreetmap.org/wiki/Relation
    """

    members: list["Relationship"]

    def __iter__(self) -> Iterator[tuple[Optional[str], Element]]:
        for relship in self.members:
            yield relship.role, relship.member


@dataclass(repr=False, eq=False)
class AreaWay(Way):
    """
    An area whose outline is defined by a closed way.

    Unless this element was produced by an ``area`` query, this is not necessarily an area
    as defined by the Overpass API, since the criteria to decide which elements represent
    an area may differ.

    Attributes:
        area_id: An ID specific to the Overpass API, which is only set when explicitly querying
                 areas. Otherwise, you cannot be sure that this object is also considered an area
                 by Overpass. You can derive this ID yourself by adding ``2_400_000_000`` onto
                 the element ID, but must consider that there is no Overpass area with that ID.
        geometry: The polygon enclosed by the way.

    References:
        - https://wiki.openstreetmap.org/wiki/Area
        - https://wiki.openstreetmap.org/wiki/Overpass_API/Areas
    """

    area_id: Optional[int]
    geometry: Optional[Polygon]


@dataclass(repr=False, eq=False)
class AreaRelation(Relation):
    """
    An area whose geometry is defined by a relation.

    Areas defined by relations may have boundaries made up of several unclosed ways.

    Relations of ``type=multipolygon`` may have boundaries ("outer" role) and holes ("inner" role)
    made up of several unclosed ways.

    Tags describing the multipolygon always go on the relation. The inner and outer ways are tagged
    if they describe something in their own right. For example,
     - a multipolygon relation may be tagged as landuse=forest if it outlines a forest,
     - its outer ways may be tagged as barrier=fence if the forest is fenced,
     - and its inner ways may be tagged as natural=water if there is a lake within the forest
       boundaries.

    Unless this element was produced by an ``area`` query, this is not necessarily an area
    as defined by the Overpass API, since the criteria to decide which elements represent
    an area may differ.

    Attributes:
        area_id: An ID specific to the Overpass API, which is only set when explicitly querying
                 areas. Otherwise, you cannot be sure that this object is also considered an area
                 by Overpass. You can derive this ID yourself by adding ``3_600_000_000`` onto the
                 element ID, but must consider that there is no Overpass area with that ID.
        geometry: The complex polygons whose boundaries and holes are made up of the ways
                  inside the relation. Members that are not ways, or are not part of any polygon
                  boundary, are not part of the result geometry. ``None`` if the geometry of the
                  relation members is not included in the query's result set.

    References:
        - https://wiki.openstreetmap.org/wiki/Area
        - https://wiki.openstreetmap.org/wiki/Relation:multipolygon
        - https://wiki.openstreetmap.org/wiki/Relation:boundary
        - https://wiki.openstreetmap.org/wiki/Overpass_API/Areas
    """

    area_id: Optional[int]
    geometry: Union[Polygon, MultiPolygon, None]


@dataclass(repr=False)
class Relationship(Spatial):
    """
    The relationship of an element that is part of a relation, with an optional role.

    Attributes:
        member:     any element
        relation:   a relation that the member is a part of
        role:       describes the function of the member in the context of the relation

    References:
        - https://wiki.openstreetmap.org/wiki/Relation#Roles
    """

    member: Element
    relation: Relation
    role: Optional[str]

    @property
    def geojson(self) -> GeoJsonDict:
        """
        A mapping of ``member``.

        This is ``member.geojson``, with the added properties ``role`` and ``__rel__``.
        """
        return _geojson_feature(self)

    def __repr__(self) -> str:
        role = f" as '{self.role}'" if self.role else " "
        return f"{type(self).__name__}({self.member}{role} in {self.relation})"


_KNOWN_ELEMENTS = {"node", "way", "relation", "area"}


_ElementKey = tuple[str, int]
"""Elements are uniquely identified by the tuple (type, id)."""

_MemberKey = tuple[_ElementKey, str]
"""Relation members are identified by their element key and role."""


class _ElementCollector:
    def __init__(self) -> None:
        self.result_set: list[_ElementKey] = []
        self.typed_dict: dict[_ElementKey, Element] = {}
        self.untyped_dict: dict[_ElementKey, OverpassDict] = defaultdict(dict)
        self.member_dict: dict[int, list[_MemberKey]] = defaultdict(list)


def collect_elements(query: Query) -> list[Element]:
    """
    Produce typed elements from the result set of a query.

    This function collects elements that are of type "node", "way", "relation", or "area".
    Derived elements with other types - f.e. produced by ``make`` and ``convert`` statements
    or when using ``out count`` - are ignored. An exception is made for areas, which are so
    closely derived from ways and relations that it makes sense to represent them as such.
    If you need to work with other derived elements, you should not use this function.

    Element data is "conflated", which means that if elements appear more than once in a
    result set, their data is merged. This is useful f.e. when querying tags for relation members:
    using ``rel(...); out tags;`` will only print tags for relation itself, not its members.
    For those you will have to recurse down from the relation, which means members will show
    up twice in the result set: once untagged as a member of the relation, and once tagged at
    the top level. This function will have these two occurrences point to the same, single object.

    The order of elements and relation members is retained.

    Args:
        query: a finished query

    Returns:
        the result set as a list of typed elements

    Raises:
        ValueError: If the input query is unfinished/has no result set.
        KeyError: The only times there should be missing keys is when either using ``out noids``,
                  or when building derived elements that are missing common keys.
    """
    if not query.done:
        msg = "query has no result set"
        raise ValueError(msg)

    collector = _ElementCollector()
    _collect_untyped(query, collector)
    _collect_typed(collector)
    _collect_relationships(collector)
    return [collector.typed_dict[elem_key] for elem_key in collector.result_set]


def _collect_untyped(query: Query, collector: _ElementCollector) -> None:
    if not query.result_set:
        raise AssertionError

    # Here we populate 'untyped_dict' with both top level elements, and
    # relation members, while conflating their data if they appear as both.
    # We also populate 'member_dict'.
    for elem_dict in query.result_set:
        if elem_dict.get("type") not in _KNOWN_ELEMENTS:
            continue

        key: _ElementKey = (elem_dict["type"], elem_dict["id"])

        collector.result_set.append(key)
        collector.untyped_dict[key].update(elem_dict)

        if elem_dict["type"] != "relation":
            continue

        for mem in elem_dict["members"]:
            key = (mem["type"], mem["ref"])
            collector.untyped_dict[key].update(mem)
            collector.member_dict[elem_dict["id"]].append((key, mem.get("role")))


def _collect_typed(collector: _ElementCollector) -> None:
    for elem_key, elem_dict in collector.untyped_dict.items():
        (elem_type, elem_id) = elem_key

        args = dict(
            id=elem_id,
            tags=elem_dict.get("tags"),
            bounds=tuple(elem_dict["bounds"].values()) if "bounds" in elem_dict else None,
            center=Point(elem_dict["center"].values()) if "center" in elem_dict else None,
            meta=Metadata(
                timestamp=elem_dict["timestamp"],
                version=elem_dict["version"],
                changeset=elem_dict["changeset"],
                user_name=elem_dict["user"],
                user_id=elem_dict["uid"],
            )
            if "timestamp" in elem_dict
            else None,
            relations=[],  # add later
        )

        cls: type[Element]

        if elem_type == "node":
            cls = Node

        elif elem_type == "way":
            cls = Way
            args["node_ids"] = elem_dict.get("nodes")
            if _is_area_element(elem_dict):
                cls = AreaWay
                args["area_id"] = elem_id % _AREA_REL_ID_OFFSET

        elif elem_type == "relation":
            cls = Relation
            args["members"] = []  # add later
            if _is_area_element(elem_dict):
                cls = AreaRelation
                args["area_id"] = elem_id % _AREA_REL_ID_OFFSET

        elif elem_id > _AREA_REL_ID_OFFSET:
            cls = AreaRelation
            args["area_id"] = elem_id % _AREA_REL_ID_OFFSET
            args["members"] = []  # add later

        else:
            cls = AreaWay
            args["area_id"] = elem_id % _AREA_WAY_ID_OFFSET

        if cls is not Relation:
            args["geometry"] = _geometry(elem_dict)

        elem = cls(**args)
        collector.typed_dict[elem_key] = elem


def _collect_relationships(collector: _ElementCollector) -> None:
    for rel_id, mem_roles in collector.member_dict.items():
        rel = cast(Relation, collector.typed_dict[("relation", rel_id)])

        for mem_key, mem_role in mem_roles:
            mem = collector.typed_dict[mem_key]
            relship = Relationship(member=mem, relation=rel, role=mem_role or None)
            mem.relations.append(relship)
            rel.members.append(relship)


# e.g. area with id 3_600_051_477 correlates with relation 51_477
_AREA_REL_ID_OFFSET = 3_600_000_000
_AREA_WAY_ID_OFFSET = 2_400_000_000


def _geometry(raw_elem: OverpassDict) -> Optional[BaseGeometry]:
    """
    Construct the geometry a given OSM element makes up.

    Args:
        raw_elem: an element from a query's result set

    Returns:
        - None if there is no geometry available for this element.
        - Point when given a node.
        - LineString when given an open way.
        - LinearRing when given a closed way, that supposedly is *not* an area.
        - Polygon when given a closed way, that supposedly is an area.
        - (Multi-)Polygons containing given a (multipolygon) relation.
          Relation members that are not ways, or are not part of any polygon boundary, are
          not part of the result geometry.

    Raises:
        ValueError: if element is not of type 'node', 'way', 'relation', or 'area'
    """
    if raw_elem.get("type") not in _KNOWN_ELEMENTS:
        msg = "expected element of type 'node', 'way', 'relation', or 'area'"
        raise ValueError(msg)

    if raw_elem["type"] == "node":
        lat, lon = raw_elem.get("lat"), raw_elem.get("lon")
        if lat and lon:
            return Point(lat, lon)

    if raw_elem["type"] == "way":
        ls = _line(raw_elem)
        if ls and ls.is_ring and _is_area_element(raw_elem):
            return Polygon(ls)
        return ls

    if _is_area_element(raw_elem):
        outers = (
            ls
            for ls in (
                _line(mem) for mem in raw_elem.get("members", ()) if mem.get("role") == "outer"
            )
            if ls
        )
        inners = (
            ls
            for ls in (
                _line(mem) for mem in raw_elem.get("members", ()) if mem.get("role") == "inner"
            )
            if ls
        )

        shells = [ls for ls in _flatten(shapely.ops.linemerge(outers)) if ls.is_closed]
        holes = [ls for ls in _flatten(shapely.ops.linemerge(inners)) if ls.is_closed]

        polys = [
            Polygon(shell=shell, holes=[hole for hole in holes if shell.contains(hole)])
            for shell in shells
        ]

        if len(polys) == 1:
            return polys[0]

        return MultiPolygon(polys)

    return None


def _line(way: OverpassDict) -> Union[LineString, LinearRing, None]:
    """Returns the geometry of a way in the result set."""
    if "geometry" not in way or len(way["geometry"]) < 2:
        return None
    is_ring = way["geometry"][0] == way["geometry"][-1]
    cls = LinearRing if is_ring else LineString
    return cls((c["lat"], c["lon"]) for c in way["geometry"])


def _flatten(obj: BaseGeometry) -> Iterable[BaseGeometry]:
    """Recursively flattens multipart geometries."""
    if isinstance(obj, BaseMultipartGeometry):
        return (nested for contained in obj.geoms for nested in _flatten(contained))
    return (obj,)


def _is_area_element(el: OverpassDict) -> bool:
    """
    Decide if ``el`` likely represents an area, and should be viewed as a (multi-)polygon.

    Args:
        el: a way or relation from a query's result set

    Returns:
        ``False`` if the input is not a relation or closed way.
        ``False``, unless there are tags which indicate that the way represents an area.

    References:
        - https://wiki.openstreetmap.org/wiki/Overpass_API/Areas
        - https://github.com/drolbr/Overpass-API/blob/master/src/rules/areas.osm3s
          (from 2018-04-09)
        - https://wiki.openstreetmap.org/wiki/Overpass_turbo/Polygon_Features
        - https://github.com/tyrasd/osm-polygon-features/blob/master/polygon-features.json
          (from 2016-11-03)
    """
    # Check if the element is explicitly an area
    if el["type"] == "area":
        return True

    # Check if a given way is open
    if el["type"] == "way" and ("geometry" not in el or el["geometry"][0] != el["geometry"][-1]):
        return False

    # Assume not an area if there are no tags available
    if "tags" not in el:
        return False

    tags = el["tags"]

    # Check if the element is explicitly tagged as not area
    if tags.get("area") == "no":
        return False

    # Check if there is a tag where any value other than 'no' suggests area
    # (note: Overpass may require the "name" tag as well)
    if any(tags.get(name, "no") != "no" for name in _AREA_TAG_NAMES):
        return True

    # Check if there are tag values that suggest area
    # (note: Overpass may require the "name" tag as well)
    return any(
        (
            v in _AREA_TAG_VALUES_ONE_OF.get(k, ())
            or v not in _AREA_TAG_VALUES_NONE_OF.get(k, (v,))
            for k, v in tags.items()
        )
    )


_AREA_TAG_NAMES = {
    "area",
    "area:highway",
    "amenity",
    "boundary",
    "building",
    "building:part",
    "craft",
    "golf",
    "historic",
    "indoor",
    "landuse",
    "leisure",
    "military",
    "office",
    "place",
    "public_transport",
    "ruins",
    "shop",
    "tourism",
    # for relations
    "admin_level",
    "postal_code",
    "addr:postcode",
}

_AREA_TAG_VALUES_ONE_OF = {
    "barrier": {"city_wall", "ditch", "hedge", "retaining_wall", "wall", "spikes"},
    "highway": {"services", "rest_area", "escape", "elevator"},
    "power": {"plant", "substation", "generator", "transformer"},
    "railway": {"station", "turntable", "roundhouse", "platform"},
    "waterway": {"riverbank", "dock", "boatyard", "dam"},
    # for relations
    "type": {"multipolygon"},
}

_AREA_TAG_VALUES_NONE_OF = {
    "aeroway": {"no", "taxiway"},
    "man_made": {"no", "cutline", "embankment", "pipeline"},
    "natural": {"no", "coastline", "cliff", "ridge", "arete", "tree_row"},
}



"""Interface for making API calls."""

import asyncio
import logging
import re
import sys
from contextlib import suppress
from dataclasses import dataclass
from typing import Optional
from urllib.parse import urljoin

from aio_overpass import __version__
from aio_overpass.error import (
    CallError,
    ClientError,
    GiveupError,
    RunnerError,
    _result_or_raise,
    _to_client_error,
)
from aio_overpass.query import DefaultQueryRunner, Query, QueryRunner

import aiohttp


__docformat__ = "google"
__all__ = (
    "Client",
    "Status",
    "DEFAULT_INSTANCE",
    "DEFAULT_USER_AGENT",
)


DEFAULT_INSTANCE = "https://overpass-api.de/api/"
DEFAULT_USER_AGENT = f"aio-overpass/{__version__} (https://github.com/timwie/aio-overpass)"


@dataclass
class Status:
    """
    Information about the API server's rate limit.

    Attributes:
        slots: The maximum number of concurrent queries per IP
               (or ``None`` if there is no rate limit).
        free_slots: The number of slots open for this IP
                    (or ``None`` if there is no rate limit).
        cooldown_secs: The number of seconds until a slot opens for this IP
                       (or 0 if there is a free slot).
        concurrency: Maximum concurrent queries configured for this client.
    """

    slots: Optional[int]
    free_slots: Optional[int]
    cooldown_secs: int
    concurrency: int

    def __repr__(self) -> str:
        f, s, c = self.free_slots, self.slots, self.cooldown_secs

        if self.slots:
            return f"{type(self).__name__}(slots={f}/{s}, cooldown={c}s)"

        return f"{type(self).__name__}(slots=∞, cooldown={c}s)"


class Client:
    """
    A client for the Overpass API.

    Requests are rate-limited according to the configured number of slots per IP for the specified
    API server. By default, queries are retried whenever the server is too busy, or the rate limit
    was exceeded. Custom query runners can be used to implement your own retry strategy.

    Args:
        url: The url of an Overpass API instance. Defaults to the main Overpass API instance.
        user_agent: A string used for the User-Agent header. It is good practice to provide a string
                    that identifies your application, and includes a way to contact you (f.e. an
                    e-mail, or a link to a repository). This is important if you make too many
                    requests, or queries that require a lot of resources.
        concurrency: Affects the maximum number of concurrent queries. Usually, the API server
                     status includes a number of slots it provides for each IP. If that is the
                     case, we pick the minimum of that number of slots and ``concurrency`` as
                     concurrency limit. If the server does not provide a limit itself,
                     ``concurrency`` will be used as concurrency limit.
        runner: You can provide another query runner if you want to implement your own retry
                strategy.

    References:
        - https://wiki.openstreetmap.org/wiki/Overpass_API#Public_Overpass_API_instances
    """

    def __init__(
        self,
        url: str = DEFAULT_INSTANCE,
        user_agent: str = DEFAULT_USER_AGENT,
        concurrency: int = 32,
        runner: Optional[QueryRunner] = None,
    ) -> None:
        if concurrency <= 0:
            msg = "'concurrency' must be > 0"
            raise ValueError(msg)

        self._url = url
        self._user_agent = user_agent
        self._concurrency = concurrency
        self._runner = runner or DefaultQueryRunner()

        self._maybe_session: Optional[aiohttp.ClientSession] = None
        self._maybe_any_status: Optional[Status] = None
        self._maybe_sem: Optional[asyncio.BoundedSemaphore] = None

    def _session(self) -> aiohttp.ClientSession:
        """The session used for all requests of this client."""
        if not self._maybe_session or self._maybe_session.closed:
            headers = {"User-Agent": self._user_agent}
            self._maybe_session = aiohttp.ClientSession(headers=headers)

        return self._maybe_session

    async def _rate_limiter(self, **kwargs) -> asyncio.BoundedSemaphore:
        """
        A rate-limiting semaphore for queries.

        This semaphore can be acquired as many times as there are query slots at the API instance.
        If all slots are acquired, further queries need to wait until a slot is released.
        """
        status = self._maybe_any_status

        if self._maybe_sem:
            return self._maybe_sem

        if not status:
            status = await self._status(**kwargs)

        self._maybe_sem = asyncio.BoundedSemaphore(status.concurrency)
        return self._maybe_sem

    async def close(self) -> None:
        """Cancel all running queries and close the underlying session."""
        if self._maybe_session and not self._maybe_session.closed:
            # do not care if this fails
            with suppress(CallError):
                _ = await self.cancel_queries()

            # is raised when there are still active queries. that's ok
            with suppress(aiohttp.ServerDisconnectedError):
                await self._maybe_session.close()

    async def _status(self, **kwargs) -> "Status":
        try:
            async with self._session().get(url=urljoin(self._url, "status"), **kwargs) as response:
                text = await response.text()
        except aiohttp.ClientError as err:
            raise _to_client_error(err) from err

        slots = 0
        free_slots = None
        cooldown_secs = 0
        concurrency = self._concurrency

        try:
            match_slots_overall = re.findall("Rate limit: (\\d+)", text)
            match_slots_available = re.findall("(\\d+) slots available now", text)
            match_cooldowns = re.findall("Slot available after: .+, in (\\d+) seconds", text)

            (slots,) = match_slots_overall
            slots = int(slots) or None

            if slots:
                cooldowns = [int(secs) for secs in match_cooldowns]

                if match_slots_available:
                    free_slots = int(match_slots_available[0])
                else:
                    free_slots = slots - len(cooldowns)

                cooldown_secs = 0 if free_slots > 0 else min(cooldowns)

                # pick the server's concurrent query limit if > 0 and < self._concurrency,
                # or self._concurrency otherwise
                concurrency = min(slots or self._concurrency, self._concurrency)
        except ValueError as err:
            raise _to_client_error(response) from err

        self._maybe_any_status = Status(
            slots=slots,
            free_slots=free_slots,
            cooldown_secs=cooldown_secs,
            concurrency=concurrency,
        )
        return self._maybe_any_status

    async def status(self) -> Status:
        """
        Check the current API status.

        Raises:
            ClientError: if the status could not be looked up
        """
        return await self._status()

    async def cancel_queries(self) -> int:
        """
        Cancel all running queries.

        This can be used to terminate runaway queries that prevent you from sending new ones.

        Returns:
            the number of terminated queries

        Raises:
            ClientError: if the request to cancel queries failed
        """
        session = self._session()
        endpoint = urljoin(self._url, "kill_my_queries")
        try:
            async with session.get(endpoint) as response:
                body = await response.text()
                killed_pids = re.findall("\\(pid (\\d+)\\)", body)
                return len(set(killed_pids))
        except aiohttp.ClientError as err:
            raise _to_client_error(err) from err

    async def run_query(self, query: Query) -> None:
        """
        Send a query to the API, and await its completion.

        "Running" the query entails acquiring a connection from the pool, waiting for a slot
        to open up, the query requests themselves (which may be retried), status requests
        when the server is busy, and cooldown periods.

        The query runner is invoked before every try.

        Args:
            query: the query to run on this API instance

        Raises:
            ClientError: when query or status requests fail. If the query was retried, the error
                         of the last try will be raised. The same exception is also captured in
                         `query.error`.
        """
        if query.done:
            return  # nothing to do

        if query.nb_tries > 0:
            query.reset()  # reset failed queries

        while True:
            await self._invoke_runner(query)
            if query.done:
                return
            await self._run_query_once(query)

    async def _invoke_runner(self, query: Query) -> None:
        try:
            await self._runner(query)
        except ClientError:
            raise
        except BaseException as err:
            raise RunnerError(err) from err

    async def _run_query_once(self, query: Query) -> None:
        logger = query.logger or logging.getLogger(f"{type(self).__module__}.{type(self).__name__}")

        if query.done:
            return

        loop = asyncio.get_event_loop()

        if query._time_start == 0.0:
            query._time_start = loop.time()

        query._time_start_try = 0.0
        query._time_end_try = 0.0

        acquired_slot = False

        try:
            await self._wait_for_slot_and_cooldown(query)
            acquired_slot = True

            query._time_start_try = loop.time()

            logger.info(f"call api for {query}")

            async with self._session().get(
                url=urljoin(self._url, "interpreter"),
                params={"data": query.code},
                timeout=_next_timeout(query),
            ) as response:
                query._time_end_try = loop.time()
                query._response = await _result_or_raise(response, query.kwargs)
                query._response_bytes = response.content.total_bytes
                query._error = None

        except aiohttp.ClientError as err:
            query._error = _to_client_error(err)

        except asyncio.TimeoutError:
            query._error = _giveup_error(query, loop)

        except ClientError as err:
            query._error = err

        finally:
            query._nb_tries += 1
            if acquired_slot:
                self._maybe_sem.release()

    async def _wait_for_slot_and_cooldown(self, query: Query) -> None:
        logger = query.logger or logging.getLogger(f"{type(self).__module__}.{type(self).__name__}")
        loop = asyncio.get_event_loop()

        # This requests an API status if we haven't done so already.
        rate_limiter = await self._rate_limiter(timeout=_next_timeout(query))

        if query._has_cooldown():
            # If this client is running too many queries, we can check the status for a
            # cooldown period. This request failing is a bit of an edge case.
            # 'query.error' will be overwritten, which means we will not check for a
            # cooldown in the next iteration.
            status = await self._status(timeout=_next_timeout(query))

            if (timeout := _next_timeout_secs(query)) and status.cooldown_secs > timeout:
                raise _giveup_error(query, loop)

            logger.info(f"{query} has cooldown for {status.cooldown_secs:.1f}s")
            await asyncio.sleep(status.cooldown_secs)

        # Limit the concurrent query requests to the number of slots available.
        if rate_limiter.locked():
            logger.info(f"{query} has to wait for a slot")

        await asyncio.wait_for(
            fut=rate_limiter.acquire(),
            timeout=_next_timeout(query).total,
        )


def _next_timeout_secs(query: Query) -> Optional[float]:
    if query.run_timeout_secs:
        # use epsilon since 0 means "no timeout"
        return max(sys.float_info.epsilon, query.run_timeout_secs - query.run_duration_secs)
    return None


def _next_timeout(query: Query) -> aiohttp.ClientTimeout:
    return aiohttp.ClientTimeout(total=_next_timeout_secs(query))


def _giveup_error(query: Query, loop: asyncio.AbstractEventLoop) -> GiveupError:
    return GiveupError(kwargs=query.kwargs, after_secs=loop.time() - query._time_start)
